import http.server
import textwrap
from itertools import groupby
from operator import itemgetter
from pathlib import Path
from typing import Dict, Union

import numpy as np
from Bio.SeqUtils import seq1
from jinja2 import Environment, FileSystemLoader

from abcfold.output.alphafold3 import AlphafoldOutput
from abcfold.output.boltz import BoltzOutput
from abcfold.output.chai import ChaiOutput
from abcfold.plots.pae_plot import create_pae_plots
from abcfold.plots.plddt_plot import plot_plddt

PORT = 8000


def get_plddt_regions(plddts: Union[np.ndarray, list]) -> dict:
    """
    Get the pLDDT regions for the model
    """
    if not isinstance(plddts, np.ndarray):
        plddts = np.array(plddts)

    regions = {}
    # replace none values with -1
    plddts = np.where(plddts is None, -1, plddts)

    v_low = np.where((0 <= plddts) & (plddts <= 50))[0]
    regions["v_low"] = get_regions_helper(v_low)
    low = np.where((plddts > 50) & (plddts < 70))[0]
    regions["low"] = get_regions_helper(low)
    confident = np.where((plddts >= 70) & (plddts < 90))[0]
    regions["confident"] = get_regions_helper(confident)
    v_confident = np.where(plddts >= 90)[0]
    regions["v_high"] = get_regions_helper(v_confident)

    return regions


def get_regions_helper(indices):
    """
    Get the regions from the indices
    """
    regions = []
    for _, g in groupby(enumerate(indices), lambda x: x[0] - x[1]):
        group = map(itemgetter(1), g)
        group = list(map(int, group))
        regions.append((group[0], group[-1]))
    return regions


def get_model_sequence_data(cif_objs) -> dict:
    """
    Get the sequence for each chain and ligand in the model, used internally
    for plotting

    Args:
        cif_objs : A list of CifFile objs

    Returns:
        dict : Chain ID and sequence data
    """
    sequence_data: dict = {}
    for cif_obj in cif_objs:
        sequence_data_ = {}
        for chain in cif_obj.get_chains():
            if cif_obj.check_ligand(chain):
                if chain.id not in sequence_data_:
                    sequence_data_[chain.id] = ""
                sequence_data_[chain.id] += "".join(
                    [atom.id[0] for residue in chain for atom in residue]
                )
            elif cif_obj.check_other(chain, ["dna"]):
                sequence_data_[chain.id] = "".join(
                    [residue.get_resname()[-1] for residue in chain]
                )
            elif cif_obj.check_other(chain, ["rna"]):
                sequence_data_[chain.id] = "".join(
                    [residue.get_resname() for residue in chain]
                )
            else:

                sequence_data_[chain.id] = "".join(
                    [seq1(residue.get_resname()) for residue in chain]
                )
        sequence_data = {
            chain_id: sorted(
                [sequence_data_[chain_id], sequence_data.get(chain_id, "")],
                reverse=True,
            )[0]
            for chain_id in sequence_data_
        }

    return sequence_data


def get_model_data(model, plot_dict, method, plddt_scores, output_dir):
    """
    Get the model data for the output page

    Args:
        model (CifFile): Model object
        plot_dict (dict): Dictionary of plots
        method (str): Method used to generate the model
        output_dir (Path): Path to the output directory
    """
    regions = get_plddt_regions(plddt_scores)
    model_data = {
        "model_id": model.name,
        "model_source": method,
        "model_path": model.pathway.as_posix(),
        "plddt_regions": regions,
        "avg_plddt": model.average_plddt,
        "h_score": model.h_score,
        "residue_clashes": model.clashes_residues,
        "atom_clashes": model.clashes,
        "pae_path": Path(plot_dict[model.pathway.as_posix()])
        .relative_to(output_dir)
        .as_posix(),
    }
    return model_data


class NoCacheHTTPRequestHandler(http.server.SimpleHTTPRequestHandler):
    def end_headers(self):
        self.send_header(
            "Cache-Control", "no-store, no-cache, must-revalidate, max-age=0"
        )
        self.send_header("Pragma", "no-cache")
        self.send_header("Expires", "0")
        super().end_headers()


def plots(outputs: list, output_dir: Path):
    """
    Generate plots for the output of the different programs

    Args:
        outputs (list): List of output objects

    """
    pathway_plots = create_pae_plots(outputs, output_dir=output_dir)
    plddt_plot_input: Dict[str, list] = get_all_cif_files(outputs)

    plot_plddt(plddt_plot_input, output_name=output_dir.joinpath("plddt_plot.html"))

    pathway_plots["plddt"] = str(output_dir.joinpath("plddt_plot.html").resolve())

    return pathway_plots


def render_template(in_file_path, out_file_path, **kwargs):
    """
    Templates the given file with the keyword arguments.

    Args:
        in_file_path (Path): The path to the template.
        out_file_path (Path): The path to output the templated file.
        **kwargs (dict): Variables to use in templating.
    """
    env = Environment(
        loader=FileSystemLoader(in_file_path.parent), keep_trailing_newline=True
    )
    template = env.get_template(in_file_path.name)
    output = template.render(**kwargs)
    with open(str(out_file_path), "w") as f:
        f.write(output)


def output_open_html_script(file_out: str, port: int = 8000):
    """
    Make a python script to open the output HTML file in the default web browser

    Args:
        file_out (str): Path to the output script
        port (int): Port to run the server on
    """

    script = f"""
    import http.server
    import socketserver
    import webbrowser
    import sys

    PORT = {port}

    class NoCacheHTTPRequestHandler(http.server.SimpleHTTPRequestHandler):
        def end_headers(self):
            self.send_header("Cache-Control",
                            "no-store, no-cache, must-revalidate, max-age=0")
            self.send_header("Pragma", "no-cache")
            self.send_header("Expires", "0")
            super().end_headers()

    try:
        with socketserver.TCPServer(("", PORT),
                                    NoCacheHTTPRequestHandler) as httpd:
            print(
                f"Serving at port {PORT}: http://localhost:{PORT}/index.html"
                )
            print("Press Ctrl+C to stop the server")
            webbrowser.open(f"http://localhost:{PORT}/index.html")
            httpd.serve_forever()
    except KeyboardInterrupt:
        print("Server stopped")
        httpd.server_close()
        sys.exit(0)
    """

    script = textwrap.dedent(script)
    with open(file_out, "w") as f:
        f.write(script)


def get_all_cif_files(outputs) -> Dict[str, list]:
    method_cif_objs: Dict[str, list] = {}

    for output in outputs:
        if isinstance(output, AlphafoldOutput):
            for seed in output.seeds:
                if "Alphafold3" not in method_cif_objs:
                    method_cif_objs["Alphafold3"] = []
                method_cif_objs["Alphafold3"].extend(output.cif_files[seed])
        elif isinstance(output, BoltzOutput):

            method_cif_objs["Boltz-1"] = output.cif_files
        elif isinstance(output, ChaiOutput):
            method_cif_objs["Chai-1"] = output.cif_files

    return method_cif_objs



from pathlib import Path

import pytest

from abcfold.html.html_utils import get_plddt_regions
from abcfold.output import file_handlers


def test_npz_file(test_data):
    test_npz = Path(test_data.test_boltz_1_6BJ9_).joinpath(
        "predictions/test_mmseqs/pae_test_mmseqs_model_1.npz"
    )
    npz_file = file_handlers.NpzFile(test_npz)

    assert isinstance(npz_file.data, dict)
    assert "pae" in npz_file.data.keys()


def test_npy_file(test_data):
    test_npy = Path(test_data.test_chai1_6BJ9_).joinpath("pae_scores.npy")
    npy_file = file_handlers.NpyFile(test_npy)

    assert npy_file.data.shape == (2, 888, 888)


def test_cif_file(test_data):
    test_cif = Path(test_data.test_alphafold3_6BJ9_).joinpath(
        "seed-1_sample-0/model.cif"
    )
    cif_file = file_handlers.CifFile(test_cif)
    cif_file.check_clashes()

    assert cif_file.clashes == 0
    assert cif_file.average_plddt == pytest.approx(95.8, rel=1e-2)
    assert cif_file.h_score == 89

    plddt_regions_check = {
        "v_low": [(0, 1), (393, 394)],
        "low": [(2, 2), (395, 395)],
        "confident": [
            (3, 3),
            (134, 139),
            (209, 211),
            (396, 396),
            (527, 532),
            (601, 604),
        ],
        "v_high": [
            (4, 133),
            (140, 208),
            (212, 392),
            (397, 526),
            (533, 600),
            (605, 787),
        ],
    }
    plddts = cif_file.residue_plddts
    plddt_regions = get_plddt_regions(plddts)

    assert plddt_regions_check == plddt_regions
    assert cif_file.name == "model"
    assert cif_file.pathway == test_cif
    assert len(cif_file.plddts) > len(
        [
            plddts
            for plddts in cif_file.get_plddt_per_residue().values()
            for plddts in plddts
        ]
    )


def test_confidence_json_file(test_data):
    confidence_json = Path(test_data.test_alphafold3_6BJ9_).joinpath(
        "seed-1_sample-0/confidences.json"
    )
    confidence_file = file_handlers.ConfidenceJsonFile(confidence_json)

    assert "atom_chain_ids" in confidence_file.data
    assert "atom_plddts" in confidence_file.data
    assert "contact_probs" in confidence_file.data
    assert "pae" in confidence_file.data



#!/usr/bin/env python

"""
Create a copy of the PAE viewer standalone HTML with the data provided
via the CLI and open it in the default browser.

The script requires a local webserver to be run with the repository root
as the webroot. You can achieve this by changing to the repository root
directory (`pae-viewer`) and using the built-in Python HTTP module:

```bash
cd /your/download/path/pae-viewer
python3 -m http.server 8000
```

This will start an HTTP server handling requests via port 8000, which
is the default for this script. However, a custom port can be set via
the `--port` option.
"""

import argparse
import json
import os
# import webbrowser
from datetime import datetime
from pathlib import Path
from typing import Any, Optional


def resolved_path(value: str) -> Path:
    path = Path(value).resolve()

    if not path.exists():
        raise argparse.ArgumentTypeError(f"File not found: {path}")

    return path


def load_pae_viewer(
    structure_path: Path,
    chain_labels: str,
    template_file: Path,
    scores_path: Optional[Path] = None,
    crosslinks_path: Optional[Path] = None,
    port: int = 8000,
    output_file: Optional[Path] = None,
):
    data = get_session_data(structure_path, chain_labels, scores_path, crosslinks_path)
    # session_path = create_session_file(
    #     structure_path.stem, data, output_file, template_file
    # )
    create_session_file(
        structure_path.stem, data, output_file, template_file
    )

    # webbrowser.open(
    #     f"localhost:{port}/{Path(*session_path.parts[-2:])}", new=0, autoraise=True
    # )


def get_session_data(
    structure_path: Path,
    chain_labels: str,
    scores_path: Optional[Path] = None,
    crosslinks_path: Optional[Path] = None,
) -> dict[str, Any]:
    data = {
        "structureFile": {
            "name": structure_path.name,
            "content": structure_path.read_text(),
        },
        "chainLabels": chain_labels,
    }

    if scores_path:
        data["scoresFile"] = {
            "name": scores_path.name,
            "content": scores_path.read_text(),
        }

    if crosslinks_path:
        data["crosslinksFile"] = {
            "name": crosslinks_path.name,
            "content": crosslinks_path.read_text(),
        }

    return data


def create_session_json_element(value: Any) -> str:
    return (
        '    <script type="application/json" id="session-data">\n'
        f"        {json.dumps(value)}\n"
        "    </script>"
    )


def create_session_file(
    handle: str, data: dict[str, str], output_file: Optional[Path], template_file: Path
) -> Path:
    output_file = Path(output_file) if output_file else None
    current_dir = Path(__file__).parent.resolve()

    timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
    random_id = os.urandom(4).hex()

    if output_file is not None:
        if not output_file.suffix == ".html":
            output_file = output_file.with_suffix(".html")
        session_path = output_file
    else:
        session_path = current_dir / f"{timestamp}_{random_id}_{handle}.html"
    session_path.touch()

    pae_template = str(template_file)
    with (
        open(pae_template, "r") as template_file_obj,
        open(session_path, "w") as session_file,
    ):
        template = template_file_obj.read()
        template = template.replace(
            "</head>", f"\n{create_session_json_element(data)}\n</head>"
        )

        session_file.write(template)

    return session_path


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawTextHelpFormatter
    )

    parser.add_argument(
        "-s",
        "--structure",
        help="path to structure file",
        type=resolved_path,
        required=True,
    )

    parser.add_argument(
        "-r",
        "--scores",
        help="path to JSON file containing scores",
        type=resolved_path,
        required=True,
    )

    parser.add_argument(
        "-l",
        "--labels",
        help="semicolon-separated list of chain labels",
        type=str,
    )

    parser.add_argument(
        "-c",
        "--crosslinks",
        help="path to CSV containing crosslinks",
        type=resolved_path,
    )

    parser.add_argument(
        "-p",
        "--port",
        help="port of the local file server",
        type=int,
        default=8000,
    )

    parser.add_argument(
        "-o",
        "--output_file",
        help="output file",
    )
    parser.add_argument(
        "-t",
        "--template_file",
        help="template file",
        type=resolved_path,
        default="pae-viewer.html",
    )

    args = parser.parse_args()

    load_pae_viewer(
        args.structure,
        args.labels,
        args.template_file,
        args.scores,
        args.crosslinks,
        args.port,
        args.output_file,
    )



import configparser
import json
import os
import shutil
import socketserver
import sys
import tempfile
import webbrowser
from pathlib import Path

from abcfold.alphafold3.run_alphafold3 import run_alphafold3
from abcfold.argparse_utils import (alphafold_argparse_util,
                                    boltz_argparse_util, chai_argparse_util,
                                    custom_template_argpase_util,
                                    main_argpase_util, mmseqs2_argparse_util,
                                    prediction_argparse_util,
                                    raise_argument_errors,
                                    visuals_argparse_util)
from abcfold.html.html_utils import (PORT, NoCacheHTTPRequestHandler,
                                     get_all_cif_files, get_model_data,
                                     get_model_sequence_data,
                                     output_open_html_script, plots,
                                     render_template)
from abcfold.output.alphafold3 import AlphafoldOutput
from abcfold.output.boltz import BoltzOutput
from abcfold.output.chai import ChaiOutput
from abcfold.output.utils import (get_gap_indicies, insert_none_by_minus_one,
                                  make_dummy_m8_file)
from abcfold.scripts.abc_script_utils import (check_input_json, make_dir,
                                              make_dummy_af3_db, setup_logger)
from abcfold.scripts.add_mmseqs_msa import add_msa_to_json

logger = setup_logger()

HTML_DIR = Path(__file__).parent / "html"
HTML_TEMPLATE = HTML_DIR.joinpath("abcfold.html.jinja2")
PLOTS_DIR = ".plots"


def run(args, config, defaults, config_file):
    """Run ABCFold

    Args:
        args (argparse.Namespace): Arguments from the command line
        config (configparser.SafeConfigParser): Config parser object
        defaults (dict): Default values from the config file
        config_file (Path): Path to the config file


    Raises:
        SystemExit: If the database directory or model parameters directory is not found


    """
    outputs = []

    args.output_dir = Path(args.output_dir)

    if args.mmseqs2:
        logger.info("MMSeqs2 Selected, all other MSAs will be ignored")

    make_dir(args.output_dir, overwrite=args.override)
    make_dir(args.output_dir.joinpath(PLOTS_DIR))

    updated_config = False
    if args.model_params is not None and args.model_params != defaults["model_params"]:
        config.set("Databases", "model_params", args.model_params)
        updated_config = True
    if args.database_dir is not None and args.database_dir != defaults["database_dir"]:
        config.set("Databases", "database_dir", args.database_dir)
        updated_config = True
    if updated_config:
        with open(config_file, "w") as f:
            config.write(f)

    args = raise_argument_errors(args)
    # Ensure that the input json file is valid
    args.input_json = check_input_json(
        args.input_json,
        output_dir=args.output_dir,
        use_af3_templates=args.use_af3_template_search,
    )

    with open(args.input_json, "r") as f:
        input_params = json.load(f)

    name = input_params.get("name")
    if name is None:
        logger.error("Input JSON must contain a 'name' field")
        sys.exit(1)

    if args.alphafold3:
        from abcfold.alphafold3.check_install import check_af3_install

        check_af3_install(interactive=False)

    if args.boltz1:
        from abcfold.boltz1.check_install import check_boltz1

        check_boltz1()

    if args.chai1:
        from abcfold.chai1.check_install import check_chai1

        check_chai1()

    with tempfile.TemporaryDirectory() as temp_dir_str:
        temp_dir = Path(temp_dir_str)
        if args.mmseqs2:
            if not args.output_json:
                input_json = Path(args.input_json)
                run_json = temp_dir.joinpath(
                    input_json.name.replace(".json", "_mmseqs.json")
                )
            else:
                run_json = Path(args.output_json)

            input_params = add_msa_to_json(
                input_json=input_json,
                templates=args.templates,
                num_templates=args.num_templates,
                chai_template_output=temp_dir.joinpath("all_chains.m8"),
                custom_template=args.custom_template,
                custom_template_chain=args.custom_template_chain,
                target_id=args.target_id,
                input_params=input_params,
                output_json=run_json,
                to_file=True,
            )

        else:
            run_json = Path(args.input_json)

        if args.alphafold3:
            af3_database = args.database_dir
            if args.mmseqs2 or (
                any(
                    seq["protein"]["unpairedMsa"] is not None
                    or seq["protein"]["unpairedMsaPath"] is not None
                    for seq in input_params["sequences"]
                )
            ):
                af3_database = make_dummy_af3_db(temp_dir)

            run_alphafold3(
                input_json=run_json,
                output_dir=args.output_dir,
                model_params=args.model_params,
                database_dir=af3_database,
                number_of_models=args.number_of_models,
                num_recycles=args.num_recycles,
            )

            # Need to find the name of the af3_dir
            af3_out_dir = list(
                [
                    file
                    for file in args.output_dir.iterdir()
                    if not file.suffix == ".json"
                ]
            )[0]
            ao = AlphafoldOutput(af3_out_dir, input_params, name)
            outputs.append(ao)
            run_json = ao.input_json

        if args.boltz1:
            from abcfold.boltz1.run_boltz import run_boltz

            run_boltz(
                input_json=run_json,
                output_dir=args.output_dir,
                save_input=args.save_input,
                number_of_models=args.number_of_models,
                num_recycles=args.num_recycles,
            )
            bolt_out_dir = list(args.output_dir.glob("boltz_results*"))[0]
            bo = BoltzOutput(bolt_out_dir, input_params, name)
            outputs.append(bo)

        if args.chai1:
            from abcfold.chai1.run_chai1 import run_chai

            template_hits_path = None
            if args.templates and args.mmseqs2:
                template_hits_path = temp_dir.joinpath("all_chain.m8")

            template_hits_path = make_dummy_m8_file(run_json, temp_dir)

            chai_output_dir = args.output_dir.joinpath("chai1")
            run_chai(
                input_json=run_json,
                output_dir=chai_output_dir,
                save_input=args.save_input,
                number_of_models=args.number_of_models,
                num_recycles=args.num_recycles,
                template_hits_path=template_hits_path,
            )

            co = ChaiOutput(chai_output_dir, input_params, name)
            outputs.append(co)

        if args.no_visuals:
            logger.info("Visuals disabled")
            return

        plot_dict = plots(outputs, args.output_dir.joinpath(PLOTS_DIR))

        # Compile data to make output page
        programs_run = []
        cif_models = [
            cif_file
            for cif_list in get_all_cif_files(outputs).values()
            for cif_file in cif_list
        ]
        indicies = get_gap_indicies(*cif_models)
        index_counter = 0

        alphafold_models = {"models": []}

        if args.alphafold3:
            programs_run.append("AlphaFold3")
            for seed in ao.output.keys():
                for idx in ao.output[seed].keys():
                    model = ao.output[seed][idx]["cif"]
                    model.check_clashes()
                    plddt = model.residue_plddts

                    plddt = insert_none_by_minus_one(indicies[index_counter], plddt)
                    index_counter += 1
                    model_data = get_model_data(
                        model, plot_dict, "AlphaFold3", plddt, args.output_dir
                    )
                    alphafold_models["models"].append(model_data)

        boltz_models = {"models": []}
        if args.boltz1:
            programs_run.append("Boltz-1")
            for idx in bo.output.keys():
                model = bo.output[idx]["cif"]
                model.check_clashes()
                plddt = model.residue_plddts
                plddt = insert_none_by_minus_one(indicies[index_counter], plddt)
                index_counter += 1
                model_data = get_model_data(
                    model, plot_dict, "Boltz-1", plddt, args.output_dir
                )
                boltz_models["models"].append(model_data)

        chai_models = {"models": []}
        if args.chai1:
            programs_run.append("Chai-1")
            for idx in co.output.keys():
                if idx >= 0:
                    model = co.output[idx]["cif"]
                    model.check_clashes()
                    plddt = model.residue_plddts
                    plddt = insert_none_by_minus_one(indicies[index_counter], plddt)
                    index_counter += 1
                    model_data = get_model_data(
                        model, plot_dict, "Chai-1", plddt, args.output_dir
                    )
                    chai_models["models"].append(model_data)

        combined_models = (
            alphafold_models["models"] + boltz_models["models"] + chai_models["models"]
        )

        sequence_data = get_model_sequence_data(cif_models)
        sequence = ""
        for key in sequence_data.keys():
            sequence += sequence_data[key]
        chain_data = {}
        ref = 0
        for key in sequence_data.keys():
            chain_data["Chain " + key] = (ref, len(sequence_data[key]) + ref - 1)
            ref += len(sequence_data[key])
        results_dict = {
            "sequence": sequence,
            "models": combined_models,
            "plotly_path": Path(plot_dict["plddt"])
            .relative_to(args.output_dir.resolve())
            .as_posix(),
            "chain_data": chain_data,
        }
        results_json = json.dumps(results_dict)

        if not args.output_dir.joinpath(".feature_viewer").exists():
            shutil.copytree(HTML_DIR, args.output_dir / ".feature_viewer")

        if len(programs_run) > 1:
            programs = (
                "Structure predictions for: "
                + ", ".join(programs_run[:-1])
                + " and "
                + programs_run[-1]
            )
        else:
            programs = "Structure predictions for: " + programs_run[0]

        # Create the index page
        HTML_OUT = args.output_dir.joinpath("index.html")
        html_out = Path(HTML_OUT).resolve()
        render_template(
            HTML_TEMPLATE,
            html_out,
            # kwargs appear as variables in the template
            abcfold_html_dir=".feature_viewer",
            programs=programs,
            results_json=results_json,
            version=0.1,
        )
        logger.info(f"Output page written to {HTML_OUT}")

        # Change to the output directory to run the server
        os.chdir(args.output_dir)

        # Make a script to open the output HTML file in the default web browser
        output_open_html_script("open_output.py", port=PORT)

        if args.no_server:
            logger.info("Server disabled")
            logger.info(
                "Run 'python open_output.py' in the output directory to \
view the output pages"
            )
            return

        try:
            # Start the server
            with socketserver.TCPServer(("", PORT), NoCacheHTTPRequestHandler) as httpd:
                logger.info(
                    f"Serving at port {PORT}: http://localhost:{PORT}/index.html"
                )
                logger.info("Press Ctrl+C to stop the server")
                # Open the main HTML page in the default web browser
                webbrowser.open(f"http://localhost:{PORT}/index.html")
                # Keep the server running
                httpd.serve_forever()
        except KeyboardInterrupt:
            logger.info("Server stopped")
            httpd.server_close()
            sys.exit(0)


def main():
    """
    Run AlphaFold3 / Boltz1 / Chai-1
    """
    import argparse

    parser = argparse.ArgumentParser(description="Run AlphaFold3 / Boltz1 / Chai-1")

    defaults = {}
    config_file = Path(__file__).parent.joinpath("data", "config.ini")
    config = configparser.SafeConfigParser()

    if config_file.exists():
        config.read(str(config_file))
        defaults.update(dict(config.items("Databases")))

    parser = main_argpase_util(parser)
    parser = alphafold_argparse_util(parser)
    parser = boltz_argparse_util(parser)
    parser = chai_argparse_util(parser)
    parser = mmseqs2_argparse_util(parser)
    parser = custom_template_argpase_util(parser)
    parser = prediction_argparse_util(parser)
    parser = visuals_argparse_util(parser)

    parser.set_defaults(**defaults)
    args = parser.parse_args()

    run(
        args,
        config,
        defaults,
        config_file,
    )


if __name__ == "__main__":
    main()



import tempfile
from pathlib import Path

from abcfold.html.html_utils import get_model_sequence_data
from abcfold.plots.pae_plot import create_pae_plots
from abcfold.plots.plddt_plot import plot_plddt


def test_plddt_plot(output_objs):

    af3_files = output_objs.af3_output.cif_files["seed-1"]
    boltz_files = output_objs.boltz_output.cif_files
    chai_files = output_objs.chai_output.cif_files

    assert len(af3_files) == len(boltz_files) == len(chai_files)
    plot_files = {
        "Alphafold3": af3_files,
        "Boltz-1": boltz_files,
        "Chai-1": chai_files,
    }

    with tempfile.TemporaryDirectory() as temp_dir:
        plot_plddt(
            plot_files,
            output_name=f"{temp_dir}/test.html",
        )

        assert Path(f"{temp_dir}/test.html").exists()


def test_pae_plots(output_objs):
    outputs = [
        output_objs.af3_output,
        output_objs.boltz_output,
        output_objs.chai_output,
    ]

    with tempfile.TemporaryDirectory() as temp_dir_str:
        temp_dir = Path(temp_dir_str)
        plot_pathways = create_pae_plots(outputs, output_dir=temp_dir)

        assert len(plot_pathways) == 6
        values = [Path(value).name for value in plot_pathways.values()]

        assert "confidences_seed-1_sample-0_af3_pae_plot.html" in values
        assert "confidences_seed-1_sample-1_af3_pae_plot.html" in values
        assert "test_mmseqs_model_0_af3_pae_pae_plot.html" in values
        assert "test_mmseqs_model_1_af3_pae_pae_plot.html" in values
        assert "pred.model_idx_0_af3_pae_pae_plot.html" in values
        assert "pred.model_idx_1_af3_pae_pae_plot.html" in values

        assert (
            "tests/test_data/alphafold3_6BJ9/seed-1_sample-0/\
model.cif"
            in plot_pathways
        )
        assert (
            "tests/test_data/alphafold3_6BJ9/seed-1_sample-1/\
model.cif"
            in plot_pathways
        )
        assert (
            "tests/test_data/boltz-1_6BJ9/predictions/test_mmseqs/\
test_mmseqs_model_0.cif"
            in plot_pathways
        )
        assert (
            "tests/test_data/boltz-1_6BJ9/predictions/test_mmseqs/\
test_mmseqs_model_1.cif"
            in plot_pathways
        )
        assert "tests/test_data/chai1_6BJ9/pred.model_idx_0.cif" in plot_pathways
        assert "tests/test_data/chai1_6BJ9/pred.model_idx_1.cif" in plot_pathways

        assert len(list(temp_dir.glob("*.html"))) == 6


def test_get_sequence_data(output_objs):
    af3_files = output_objs.af3_output.cif_files["seed-1"]
    boltz_files = output_objs.boltz_output.cif_files
    chai_files = output_objs.chai_output.cif_files

    cif_files = []

    [cif_files.extend(files) for files in [af3_files, boltz_files, chai_files]]

    outputdic = get_model_sequence_data(cif_files)

    assert outputdic == {
        "A": "GTGSRPITDVVFVGAARTPIGSFRSAFNNVPVTVLGREALKGALKNANVKPSLVQEAFIGVVVPSNAGQGPA\
RQVVLGAGCDVSTVVTAVNKMCASGMKAIACAASILQLDLQEMVVAGGMESMSCVPFYLPRGEIPFGGTKLIDGIPRDGLNDVYND\
ILMGACADKVAKQFAITREEQDKYAILSYKRSAAAWKEGIFAKEIIPLEVTQGKKTITVEEDEEYKKVNFEKIPKLKPAFTSEGSV\
TAANASTLNDGAAMVVMTTVDGAKKHGLKPLARMLAYGDAATHPIDFGIAPASVIPKVLKLAGLQIKDIDLWEINEAFAVVPLYTM\
KTLGLDESKVNIHGGAVSLGHPIGMSGARIVGHLVHTLKPGQKGCAAICNGGGGAGGMIIEKL",
        "B": "GTGSRPITDVVFVGAARTPIGSFRSAFNNVPVTVLGREALKGALKNANVKPSLVQEAFIGVVVPSNAGQGPA\
RQVVLGAGCDVSTVVTAVNKMCASGMKAIACAASILQLDLQEMVVAGGMESMSCVPFYLPRGEIPFGGTKLIDGIPRDGLNDVYND\
ILMGACADKVAKQFAITREEQDKYAILSYKRSAAAWKEGIFAKEIIPLEVTQGKKTITVEEDEEYKKVNFEKIPKLKPAFTSEGSV\
TAANASTLNDGAAMVVMTTVDGAKKHGLKPLARMLAYGDAATHPIDFGIAPASVIPKVLKLAGLQIKDIDLWEINEAFAVVPLYTM\
KTLGLDESKVNIHGGAVSLGHPIGMSGARIVGHLVHTLKPGQKGCAAICNGGGGAGGMIIEKL",
        "C": "NCNCCCNNCNCCOCOPOOOCOCOPOOOPOOOCCCCCOCONCCCONCCSCOC",
        "D": "NCNCCCNNCNCCOCOPOOOCOCOPOOOPOOOCCCCCOCONCCCONCCSCOC",
    }



# import numpy as np
import logging
from pathlib import Path
from typing import Any, Dict, List, Union

import plotly.express as px
import plotly.graph_objects as go
import plotly.offline as pyo

from abcfold.output.file_handlers import CifFile
from abcfold.output.utils import get_gap_indicies, insert_none_by_minus_one

logger = logging.getLogger("logger")


def plot_plddt(
    cif_models_dict: Dict[str, List[CifFile]],
    output_name: Union[str, Path],
    line_width: float = 1.6,
    dash: str = "dot",
    show: bool = False,
    chain_line_occupancy: float = 0.9,
    include_plotlyjs: bool = True,
) -> None:
    """
    Plots the pLDDT distribution of the models in the dictionary of cif models. Outputs
    and html file with the plot.

    Args:
        cif_models_dict: Dictionary of cif models to plot. The keys are the source of
            the models and the values are lists of CifFile objects.
            e.g. {"Alphafold3": [CifFile, CifFile, ...], "Boltz-1": [CifFile, ...],
            "Chai-1": [CifFile, ...]}

        output_name: Path to the output html file.
        line_width: Width of the lines in the plot.
        dash: Dash style of the lines in the plot.
        show: If True, the plot will be displayed in the browser.
        chain_line_occupancy: Opacity of the vertical lines that separate the chains.

    Returns:
        None

    Outputs:
        An html file with the plot.
    """

    fig = go.Figure()
    fig.update_xaxes(showgrid=True, gridwidth=1, gridcolor="LightGrey")
    fig.update_yaxes(showgrid=True, gridwidth=1, gridcolor="LightGrey")

    colours = list(px.colors.qualitative.T10)
    method_colours = {
        "Alphafold3": px.colors.qualitative.Set1,
        "Boltz-1": px.colors.qualitative.Set2,
        "Chai-1": px.colors.qualitative.Prism,
    }

    line_ranges: dict = {}

    cif_models = [
        cif_file for cif_files in cif_models_dict.values() for cif_file in cif_files
    ]
    indicies = get_gap_indicies(*cif_models)
    indicies_index = 0

    for method, cif_models in cif_models_dict.items():

        for cif_model in cif_models:
            model_index = int(cif_model.name.split("_")[-1])
            color_list = method_colours.get(method, colours)
            color = color_list[model_index % len(color_list)]

            plddt = cif_model.residue_plddts

            plddt = insert_none_by_minus_one(indicies[indicies_index], plddt)

            indicies_index += 1
            chain_ranges = {
                chain: len(plddt)
                for chain, plddt in cif_model.get_plddt_per_residue().items()
            }
            line_ranges = {
                chain: max(chain_ranges[chain], line_ranges.get(chain, 0))
                for chain in chain_ranges
            }

            trace = go.Scatter(
                x=list(range(len(plddt))),
                y=plddt,
                mode="lines",
                legendgroup=method,
                legendgrouptitle_text=Bold(method),
                name=f"Model {model_index + 1}",
                line=dict(dash=dash, width=line_width, color=color),
                visible=True,  # Ensure traces start as visible
                showlegend=True,
            )
            fig.add_trace(trace)

    counter = 0
    colour_index = 0
    for chain, chain_range in line_ranges.items():
        counter += chain_range
        chain_name = f"Chain {chain}"

        fig.add_vline(
            x=counter - 1,
            line=dict(color=colours[colour_index % len(colours)], dash="dash"),
            opacity=chain_line_occupancy,
            annotation_text=Bold(chain_name),
            annotation_font_size=15,
            annotation_position="top left",
            annotation_textangle=-90,
        )

        colour_index += 1
    # Create buttons for each model
    buttons = []
    num_models = len(cif_models_dict[next(iter(cif_models_dict))])

    # Add a button to show all traces
    buttons.append(
        dict(
            method="update",
            args=[{"visible": [True] * len(fig.data)}, {"showlegend": True}],
            label="All",
        )
    )

    # Add buttons for each individual model
    for model_index in range(num_models):
        button: Dict[str, Any] = dict(
            method="update",
            args=[
                {
                    "visible": [
                        i % num_models == model_index for i in range(len(fig.data))
                    ]
                },
                {"showlegend": True},
            ],
            label=f"Model {model_index + 1}",
        )
        for i in range(model_index, len(fig.data), num_models):
            button["args"][0]["visible"][i] = True
        buttons.append(button)

    # Add the updatemenu to the layout
    fig.update_layout(
        updatemenus=[
            dict(
                type="buttons",
                showactive=True,
                buttons=buttons,
                direction="left",
                x=0.5,
                xanchor="center",
                y=-0.1,
                yanchor="top",
            )
        ],
        xaxis_title=Bold("Residue Number"),
        yaxis_title=Bold("pLDDT Score"),
        title=Bold("pLDDT Distribution"),
        plot_bgcolor="white",
    )

    if show:
        fig.show()

    output_name = Path(output_name)

    if output_name.suffix == "":
        output_name = output_name.with_suffix(".html")

    if include_plotlyjs:
        fig.write_html(str(output_name))
    else:
        div = pyo.plot(fig, include_plotlyjs=False, output_type="div")
        output_name = output_name.with_suffix(".div.html")
        with open(output_name, "w") as f:
            f.write(div)


def Bold(string):
    return f"<b>{string}</b>"



import subprocess


def test_alphafold3_script():
    with subprocess.Popen(
        "python -m abcfold.abcfold --help",
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as p:
        stdout, stderr = p.communicate()
        print(stdout)
        print(stderr)
        assert p.returncode == 0, f"Return code: {p.returncode}\n{stderr.decode()}"
        assert (
            b"usage: abcfold" in stdout
        ), f"stdout: {stdout.decode()}\nstderr: {stderr.decode()}"


def test_add_custom_template_script():
    with subprocess.Popen(
        "python -m abcfold.scripts.add_custom_template --help",
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as p:
        stdout, stderr = p.communicate()
        assert p.returncode == 0, f"Return code: {p.returncode}\n{stderr.decode()}"
        assert (
            b"usage: add_custom_template" in stdout
        ), f"stdout: {stdout.decode()}\nstderr: {stderr.decode()}"


def test_add_mmseqs_msa_script():
    with subprocess.Popen(
        "python -m abcfold.scripts.add_mmseqs_msa --help",
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as p:
        stdout, stderr = p.communicate()
        assert p.returncode == 0, f"Return code: {p.returncode}\n{stderr.decode()}"
        assert (
            b"usage: add_mmseqs_msa" in stdout
        ), f"stdout: {stdout.decode()}\nstderr: {stderr.decode()}"



import os
import subprocess
from pathlib import Path

import pytest

from abcfold.alphafold3.run_alphafold3 import generate_af3_cmd, run_alphafold3


def test_generate_af3_command(test_data):
    input_json = Path(test_data.test_inputA_json)
    output_dir = Path("/road/to/nowhere")
    model_params = Path("/road/to/nowhere")
    database_dir = Path("/road/to/nowhere")

    cmd = generate_af3_cmd(
        input_json=input_json,
        output_dir=output_dir,
        model_params=model_params,
        database_dir=database_dir,
        interactive=True,
    )
    assert "docker run -it" in cmd
    assert f"--volume {input_json.parent.resolve()}:/root/af_input" in cmd
    assert f"--volume {output_dir.resolve()}:/root/af_output" in cmd
    assert f"--volume {model_params}:/root/models" in cmd
    assert f"--volume {database_dir}:/root/public_databases" in cmd
    assert "--gpus all" in cmd
    assert "alphafold3" in cmd
    assert f"--json_path=/root/af_input/{input_json.name}" in cmd
    assert "--model_dir=/root/models" in cmd
    assert "--output_dir=/root/af_output" in cmd

    cmd = generate_af3_cmd(
        input_json=input_json,
        output_dir=output_dir,
        model_params=model_params,
        database_dir=database_dir,
        interactive=False,
    )
    assert "docker run" in cmd


@pytest.mark.skipif(os.getenv("CI") == "true", reason="Skipping test in CI environment")
def test_run_af3(test_data):
    input_json = Path(test_data.test_inputA_json)
    output_dir = Path("/road/to/nowhere")
    model_params = Path("/road/to/nowhere")
    database_dir = Path("/road/to/nowhere")

    with pytest.raises(subprocess.CalledProcessError):
        run_alphafold3(
            input_json,
            output_dir,
            model_params,
            database_dir,
            interactive=False,
        )

    cmd = generate_af3_cmd(
        input_json=input_json,
        output_dir=output_dir,
        model_params=model_params,
        database_dir=database_dir,
        interactive=False,
    )

    with subprocess.Popen(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ) as p:
        _, stderr = p.communicate()
        assert p.returncode == 126
        print(stderr.decode())
        assert (
            "Error response from daemon: error while creating mount source path \
'/road/to/nowhere': mkdir /road: permission denied"
            in stderr.decode()
        )



#!/usr/bin/env python

import json
import logging
import os
import random
import tarfile
import tempfile
import time
from io import StringIO
from typing import Sequence

import pandas as pd
import requests  # type: ignore
from tqdm.autonotebook import tqdm

from abcfold.argparse_utils import (custom_template_argpase_util,
                                    mmseqs2_argparse_util)
from abcfold.scripts.abc_script_utils import (align_and_map,
                                              extract_sequence_from_mmcif,
                                              get_custom_template, get_mmcif)

logger = logging.getLogger("logger")

TQDM_BAR_FORMAT = (
    "{l_bar}{bar}| {n_fmt}/{total_fmt} [elapsed: {elapsed} remaining: {remaining}]"
)


class MMseqs2Exception(Exception):
    def __init__(self):

        msg = "MMseqs2 API is giving errors. Please confirm your input is a valid \
protein sequence. If error persists, please try again an hour later."
        logger.error(msg)
        super().__init__()


def add_msa_to_json(
    input_json,
    templates,
    num_templates,
    chai_template_output,
    custom_template,
    custom_template_chain,
    target_id,
    input_params=None,
    output_json=None,
    to_file=True,
):
    if input_params is None:
        with open(input_json, "r") as f:
            input_params = json.load(f)

    for sequence in input_params["sequences"]:
        if "protein" in sequence:
            input_id = sequence["protein"]["id"]
            input_sequence = sequence["protein"]["sequence"]
            with tempfile.TemporaryDirectory() as tmpdir:
                logger.info(f"Running MMseqs2 on sequence: {input_sequence}")
                # Run MMseqs2 to get unpaired MSA
                if templates:
                    a3m_lines, templates = run_mmseqs(
                        input_sequence,
                        tmpdir,
                        use_templates=True,
                        num_templates=num_templates,
                    )

                    for i in input_id:
                        table = pd.read_csv(
                            f"{tmpdir}/pdb70.m8",
                            delimiter="\t",
                            header=None,
                            names=[
                                "query_id",
                                "subject_id",
                                "pident",
                                "length",
                                "mismatch",
                                "gapopen",
                                "query_start",
                                "query_end",
                                "subject_start",
                                "subject_end",
                                "evalue",
                                "bitscore",
                                "comment",
                            ],
                        )

                        table["query_id"] = i

                        if chai_template_output:
                            if os.path.exists(chai_template_output):
                                table.to_csv(
                                    chai_template_output,
                                    sep="\t",
                                    index=False,
                                    header=False,
                                    mode="a",
                                )
                            else:
                                table.to_csv(
                                    chai_template_output,
                                    sep="\t",
                                    index=False,
                                    header=False,
                                )

                else:
                    a3m_lines = run_mmseqs(input_sequence, tmpdir, use_templates=False)
                    templates = []

                if custom_template:
                    for template in custom_template:
                        if not os.path.exists(template):
                            msg = f"Custom template file {template} not found"
                            logger.critical(msg)
                            raise FileNotFoundError()
                        # Can only add templates to protein sequences, so check if there
                        # are multiple protein sequences in the input json
                        if (
                            len(
                                [
                                    x
                                    for x in input_params["sequences"]
                                    if "protein" in x.keys()
                                ]
                            )
                            > 1
                            and not target_id
                        ):
                            msg = "Multiple sequences found in input json. \
Please specify target id so that custom template can be added to the correct sequence"
                            raise ValueError(msg)

                    if target_id and len(target_id) > 1:
                        if (len(custom_template) != len(target_id)) or (
                            len(custom_template_chain) != len(target_id)
                        ):
                            msg = "If providing templates for multiple targets, the \
number of target ids must match the number of custom templates and custom template \
chains"
                            raise ValueError(msg)
                        custom_templates = zip(
                            target_id, custom_template, custom_template_chain
                        )
                    else:
                        if len(custom_template) != len(custom_template_chain):
                            msg = "Number of custom templates must match the number of \
custom template chains"
                            raise ValueError(msg)
                        # if a single target id is provided, assume all custom templates
                        # are for the same target
                        if target_id:
                            target_ids = [target_id[0]] * len(custom_template)
                        else:
                            target_ids = [None] * len(custom_template)
                        custom_templates = zip(
                            target_ids, custom_template, custom_template_chain
                        )

                    for i in custom_templates:
                        tid, c_tem, c_tem_chn = i
                        sequence = get_custom_template(
                            sequence,
                            tid,
                            c_tem,
                            c_tem_chn,
                        )

                # Add unpaired MSA to the json
                sequence["protein"]["unpairedMsa"] = a3m_lines[0]
                sequence["protein"]["pairedMsa"] = ""
                sequence["protein"]["templates"] = templates

    if to_file:
        if output_json:
            with open(output_json, "w") as f:
                json.dump(input_params, f)
        else:
            output_json = input_json.replace(".json", "_mmseqs.json")
            with open(output_json, "w") as f:
                json.dump(input_params, f)

    return input_params


# Code from https://github.com/sokrypton/ColabFold
def run_mmseqs(
    x,
    prefix,
    use_env=True,
    use_filter=True,
    use_templates=False,
    filter=None,
    use_pairing=False,
    host_url="https://a3m.mmseqs.com",
    num_templates=20,
) -> Sequence[object]:
    submission_endpoint = "ticket/pair" if use_pairing else "ticket/msa"

    def submit(seqs, mode, N=101):
        n, query = N, ""
        for seq in seqs:
            query += f">{n}\n{seq}\n"
            n += 1

        res = requests.post(
            f"{host_url}/{submission_endpoint}", data={"q": query, "mode": mode}
        )
        try:
            out = res.json()
        except ValueError:
            logger.error(f"Server didn't reply with json: {res.text}")
            out = {"status": "ERROR"}
        return out

    def status(ID):
        res = requests.get(f"{host_url}/ticket/{ID}")
        try:
            out = res.json()
        except ValueError:
            logger.error(f"Server didn't reply with json: {res.text}")
            out = {"status": "ERROR"}
        return out

    def download(ID, path):
        res = requests.get(f"{host_url}/result/download/{ID}")
        with open(path, "wb") as out:
            out.write(res.content)

    # process input x
    seqs = [x] if isinstance(x, str) else x

    # compatibility to old option
    if filter is not None:
        use_filter = filter

    # setup mode
    if use_filter:
        mode = "env" if use_env else "all"
    else:
        mode = "env-nofilter" if use_env else "nofilter"

    if use_pairing:
        mode = ""
        use_templates = False
        use_env = False

    # define path
    path = prefix
    if not os.path.isdir(path):
        os.mkdir(path)

    # call mmseqs2 api
    tar_gz_file = f"{path}/out.tar.gz"
    N, REDO = 101, True

    # deduplicate and keep track of order
    seqs_unique = list(set(seqs))
    Ms = [N + seqs_unique.index(seq) for seq in seqs]
    # lets do it!
    if not os.path.isfile(tar_gz_file):
        TIME_ESTIMATE = 150 * len(seqs_unique)
        with tqdm(total=TIME_ESTIMATE, bar_format=TQDM_BAR_FORMAT) as pbar:
            while REDO:
                pbar.set_description("SUBMIT")

                # Resubmit job until it goes through
                out = submit(seqs_unique, mode, N)
                while out["status"] in ["UNKNOWN", "RATELIMIT"]:
                    sleep_time = 5 + random.randint(0, 5)
                    logger.info(f"Sleeping for {sleep_time}s. Reason: {out['status']}")
                    # resubmit
                    time.sleep(sleep_time)
                    out = submit(seqs_unique, mode, N)

                if out["status"] == "ERROR":
                    raise MMseqs2Exception()

                if out["status"] == "MAINTENANCE":
                    raise MMseqs2Exception()

                # wait for job to finish
                ID, TIME = out["id"], 0
                pbar.set_description(out["status"])
                while out["status"] in ["UNKNOWN", "RUNNING", "PENDING"]:
                    t = 5 + random.randint(0, 5)
                    logger.info(f"Sleeping for {t}s. Reason: {out['status']}")
                    time.sleep(t)
                    out = status(ID)
                    pbar.set_description(out["status"])
                    if out["status"] == "RUNNING":
                        TIME += t
                        pbar.update(n=t)

                if out["status"] == "COMPLETE":
                    if TIME < TIME_ESTIMATE:
                        pbar.update(n=(TIME_ESTIMATE - TIME))
                    REDO = False

                if out["status"] == "ERROR":
                    REDO = False
                    raise MMseqs2Exception()

            # Download results
            download(ID, tar_gz_file)

    # prep list of a3m files
    if use_pairing:
        a3m_files = [f"{path}/pair.a3m"]
    else:
        a3m_files = [f"{path}/uniref.a3m"]
        if use_env:
            a3m_files.append(f"{path}/bfd.mgnify30.metaeuk30.smag30.a3m")

    # extract a3m files
    if any(not os.path.isfile(a3m_file) for a3m_file in a3m_files):
        with tarfile.open(tar_gz_file) as tar_gz:
            tar_gz.extractall(path)

    # gather a3m lines
    a3m_lines: dict = {}
    for a3m_file in a3m_files:
        update_M, M = True, None
        for line in open(a3m_file, "r"):
            if len(line) > 0:
                if "\x00" in line:
                    line = line.replace("\x00", "")
                    update_M = True
                if line.startswith(">") and update_M:
                    M = int(line[1:].rstrip())
                    update_M = False
                    if M not in a3m_lines:
                        a3m_lines[M] = []
                a3m_lines[M].append(line)

    a3m_lines_list = ["".join(a3m_lines[n]) for n in Ms]

    tested_pdbs = []
    templates = []
    if use_templates:
        logger.info("Finding and preparing templates")
        count = 0
        for line in open(f"{path}/pdb70.m8", "r"):
            template = {}
            if count < num_templates:
                p = line.rstrip().split()
                pdb, qid, alilen, tstart, tend = (
                    p[1],
                    float(p[2]),
                    float(p[3]),
                    int(p[8]),
                    int(p[9]),
                )
                coverage = alilen / len(x)
                pdb_id = pdb.split("_")[0]

                # Use the same template filters as AF3 and only use 1 template per PDB
                if (
                    qid == 1.0
                    and coverage >= 0.95
                    or coverage < 0.1
                    or pdb_id in tested_pdbs
                ):
                    continue

                pdb_id = pdb.split("_")[0]
                cif_str = fetch_mmcif(pdb_id, pdb.split("_")[1], tstart, tend, prefix)
                template["mmcif"] = cif_str

                template_seq = extract_sequence_from_mmcif(StringIO(cif_str))
                query_indices, template_indices = align_and_map(x, template_seq)

                template["queryIndices"] = query_indices
                template["templateIndices"] = template_indices
                templates.append(template)
                tested_pdbs.append(pdb_id)
                count += 1
        logger.info(f"Found the following templates: {tested_pdbs}")

    return (a3m_lines_list, templates) if use_templates else a3m_lines_list


def fetch_mmcif(
    pdb_id,
    chain_id,
    start,
    end,
    tmpdir,
):
    """
    Fetch the mmcif file for a given PDB ID
    and chain ID and prepare it for use in AlphaFold3
    """
    pdb_id = pdb_id.lower()
    url_base = "http://www.ebi.ac.uk/pdbe-srv/view/files/"
    url = url_base + pdb_id + ".cif"
    response = requests.get(url)
    text = response.text

    output = os.path.join(tmpdir, pdb_id + ".cif")
    with open(output, "w") as f:
        f.write(text)

    return get_mmcif(output, pdb_id, chain_id, start, end, tmpdir)


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="Add MMseqs2 unpaired MSA to AlphaFold3 json"
    )
    parser.add_argument("--input_json", help="Input alphafold3 json file")
    parser.add_argument("--output_json", help="Output alphafold3 json file")

    parser = mmseqs2_argparse_util(parser)
    parser = custom_template_argpase_util(parser)

    args = parser.parse_args()

    add_msa_to_json(  # pragma: no cover
        args.input_json,
        args.templates,
        args.num_templates,
        args.custom_template,
        args.custom_template_chain,
        args.target_id,
        output_json=args.output_json,
        to_file=True,
    )


if __name__ == "__main__":  # pragma: no cover
    main()



import logging
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Union

from abcfold.chai1.af3_to_chai import ChaiFasta
from abcfold.chai1.check_install import check_chai1

logger = logging.getLogger("logger")


def run_chai(
    input_json: Union[str, Path],
    output_dir: Union[str, Path],
    save_input: bool = False,
    test: bool = False,
    number_of_models: int = 5,
    num_recycles: int = 10,
    use_templates_server: bool = False,
    template_hits_path: Path | None = None,
) -> None:
    """
    Run Chai-1 using the input JSON file

    Args:
        input_json (Union[str, Path]): Path to the input JSON file
        output_dir (Union[str, Path]): Path to the output directory
        save_input (bool): If True, save the input fasta file and MSA to the output
        directory
        test (bool): If True, run the test command
        number_of_models (int): Number of models to generate
        num_recycles (int): Number of trunk recycles
        use_templates_server (bool): If True, use templates from the server
        template_hits_path (Path): Path to the template hits m8 file

    Returns:
        None

    """
    input_json = Path(input_json)
    output_dir = Path(output_dir)
    logger.debug("Checking if Chai-1 is installed")
    check_chai1()

    with tempfile.TemporaryDirectory() as temp_dir:
        working_dir = Path(temp_dir)
        if save_input:
            logger.info("Saving input fasta file and msa to the output directory")
            working_dir = output_dir

        chai_fasta = ChaiFasta(working_dir)
        chai_fasta.json_to_fasta(input_json)

        out_fasta = chai_fasta.fasta
        msa_dir = chai_fasta.working_dir
        out_constraints = chai_fasta.constraints

        cmd = (
            generate_chai_command(
                out_fasta,
                msa_dir,
                out_constraints,
                output_dir,
                number_of_models,
                num_recycles=num_recycles,
                use_templates_server=use_templates_server,
                template_hits_path=template_hits_path,
            )
            if not test
            else generate_chai_test_command()
        )

        logger.info("Running Chai-1")
        with subprocess.Popen(
            cmd,
            stdout=sys.stdout,
            stderr=subprocess.PIPE,
        ) as proc:
            _, stderr = proc.communicate()
            if proc.returncode != 0:
                if proc.stderr:
                    logger.error(stderr.decode())
                raise subprocess.CalledProcessError(proc.returncode, proc.args)

        logger.info("Chai-1 run complete")


def generate_chai_command(
    input_fasta: Union[str, Path],
    msa_dir: Union[str, Path],
    input_constraints: Union[str, Path],
    output_dir: Union[str, Path],
    number_of_models: int = 5,
    num_recycles: int = 10,
    use_templates_server: bool = False,
    template_hits_path: Path | None = None,
) -> list:
    """
    Generate the Chai-1 command

    Args:
        input_fasta (Union[str, Path]): Path to the input fasta file
        msa_dir (Union[str, Path]): Path to the MSA directory
        input_constraints (Union[str, Path]): Path to the input constraints file
        output_dir (Union[str, Path]): Path to the output directory
        number_of_models (int): Number of models to generate
        num_recycles (int): Number of trunk recycles
        use_templates_server (bool): If True, use templates from the server
        template_hits_path (Path): Path to the template hits m8 file

    Returns:
        list: The Chai-1 command

    """

    chai_exe = Path(__file__).parent / "chai.py"
    cmd = ["python", str(chai_exe), "fold", str(input_fasta)]

    if Path(msa_dir).exists():
        cmd += ["--msa-directory", str(msa_dir)]
    if Path(input_constraints).exists():
        cmd += ["--constraint-path", str(input_constraints)]

    cmd += ["--num-diffn-samples", str(number_of_models)]
    cmd += ["--num-trunk-recycles", str(num_recycles)]

    assert not (
        use_templates_server and template_hits_path
    ), "Cannot specify both templates server and path"

    if shutil.which("kalign") is None and (use_templates_server or template_hits_path):
        logger.warning(
            "kalign not found, skipping template search kalign is required. \
Please install kalign to use templates with Chai-1."
        )
    else:
        if use_templates_server:
            cmd += ["--use-templates-server"]
        if template_hits_path:
            cmd += ["--template-hits-path", str(template_hits_path)]

    cmd += [str(output_dir)]

    return cmd


def generate_chai_test_command() -> list:
    """
    Generate the Chai-1 test command

    Args:
        None

    Returns:
        list: The Chai-1 test command
    """
    chai_exe = Path(__file__).parent / "chai.py"
    return [
        "python",
        chai_exe,
        "fold",
        "--help",
    ]



import json
from itertools import zip_longest
from pathlib import Path
from typing import List, Union

import numpy as np
import pandas as pd
from Bio.Align import PairwiseAligner

from abcfold.output.file_handlers import CifFile

AF3TEMPLATE: dict = {
    "atom_chain_ids": [],
    "atom_plddts": [],
    "contact_probs": [],
    "pae": [],
    "token_chain_ids": [],
    "token_res_ids": [],
}


class Af3Pae:
    @classmethod
    def from_alphafold3(cls, scores: dict, cif_file: CifFile):
        def reorder_matrix(pae_matrix, chain_lengths, af3_chain_lengths):
            if not isinstance(pae_matrix, np.ndarray):
                pae_matrix = np.array(pae_matrix)
            desired_order = flatten(
                [[key] * value for key, value in chain_lengths.items()]
            )
            current_order = flatten(
                [[key] * value for key, value in af3_chain_lengths.items()]
            )

            order_mapping = {}

            for i, chain_id in enumerate(current_order):
                if chain_id in desired_order:
                    order_mapping[i] = desired_order.index(chain_id)
                    desired_order[desired_order.index(chain_id)] = None  # Mark as used
                else:
                    order_mapping[i] = i

            # Reorder the PAE matrix rows and columns based on the mapping

            reordered_matrix = np.zeros_like(pae_matrix)
            for i in range(len(pae_matrix)):
                for j in range(len(pae_matrix)):

                    reordered_matrix[order_mapping[i], order_mapping[j]] = pae_matrix[
                        i, j
                    ]

            return reordered_matrix.tolist()

        af3_scores = AF3TEMPLATE.copy()

        chain_lengths = cif_file.chain_lengths(
            mode="residues", ligand_atoms=True, ptm_atoms=True
        )

        af3pae_chain_lengths = {
            k: chain_lengths[k] for k in np.unique(scores["token_chain_ids"])
        }
        if list(chain_lengths.keys()) == list(af3pae_chain_lengths.keys()):
            return cls(scores)

        residue_lengths = cif_file.chain_lengths(mode="all", ligand_atoms=True)

        atom_chain_ids = flatten(
            [[key] * value for key, value in residue_lengths.items()]
        )
        atom_plddts = cif_file.plddts

        token_res_ids_dict = cif_file.token_residue_ids()
        token_res_ids = flatten([value for _, value in token_res_ids_dict.items()])

        reordered_pae = reorder_matrix(
            scores["pae"], chain_lengths, af3pae_chain_lengths
        )
        contact_probs = reorder_matrix(
            scores["contact_probs"], chain_lengths, af3pae_chain_lengths
        )
        token_chain_ids = flatten(
            [[key] * len(value) for key, value in token_res_ids_dict.items()]
        )

        assert len(atom_chain_ids) == len(scores["atom_chain_ids"])
        assert sum([len(value) for value in token_res_ids_dict.values()]) == len(
            scores["pae"]
        )
        assert len(atom_plddts) == len(scores["atom_plddts"])

        assert len(reordered_pae) == len(scores["pae"])
        assert len(contact_probs) == len(scores["contact_probs"])
        assert len(reordered_pae[0]) == len(scores["pae"][0])
        assert len(contact_probs[0]) == len(scores["contact_probs"][0])
        assert len(token_chain_ids) == len(scores["token_chain_ids"])
        assert len(token_res_ids) == len(scores["token_res_ids"])

        af3_scores["atom_chain_ids"] = atom_chain_ids
        af3_scores["atom_plddts"] = atom_plddts
        af3_scores["contact_probs"] = contact_probs
        af3_scores["pae"] = reordered_pae
        af3_scores["token_chain_ids"] = token_chain_ids
        af3_scores["token_res_ids"] = token_res_ids

        return cls(af3_scores)

    @classmethod
    def from_boltz1(cls, scores: dict, cif_file: CifFile):
        af3_scores = AF3TEMPLATE.copy()

        chain_lengths = cif_file.chain_lengths(mode="residues", ligand_atoms=True)
        residue_lengths = cif_file.chain_lengths(mode="all", ligand_atoms=True)

        atom_chain_ids = flatten(
            [[key] * value for key, value in residue_lengths.items()]
        )

        atom_plddts = cif_file.plddts
        token_chain_ids = flatten(
            [[key] * value for key, value in chain_lengths.items()]
        )

        token_res_ids = flatten(
            [
                [value for value in values]
                for _, values in cif_file.token_residue_ids().items()
            ]
        )

        af3_scores["pae"] = scores["pae"].tolist()
        af3_scores["atom_chain_ids"] = atom_chain_ids
        af3_scores["atom_plddts"] = atom_plddts
        af3_scores["contact_probs"] = np.zeros(shape=scores["pae"].shape).tolist()
        af3_scores["token_chain_ids"] = token_chain_ids
        af3_scores["token_res_ids"] = token_res_ids

        return cls(af3_scores)

    @classmethod
    def from_chai1(cls, scores: np.ndarray, cif_file: CifFile):
        af3_scores = AF3TEMPLATE.copy()
        chain_lengths = cif_file.chain_lengths(mode="residues", ligand_atoms=True)

        residue_lengths = cif_file.chain_lengths(mode="all", ligand_atoms=True)

        atom_chain_ids = flatten(
            [[key] * value for key, value in residue_lengths.items()]
        )

        atom_plddts = cif_file.plddts
        token_chain_ids = flatten(
            [[key] * value for key, value in chain_lengths.items()]
        )

        token_res_ids = flatten(
            [
                [value for value in values]
                for _, values in cif_file.token_residue_ids().items()
            ]
        )

        af3_scores["pae"] = scores.tolist()
        af3_scores["atom_chain_ids"] = atom_chain_ids
        af3_scores["atom_plddts"] = atom_plddts
        af3_scores["contact_probs"] = np.zeros(shape=scores.shape).tolist()
        af3_scores["token_chain_ids"] = token_chain_ids
        af3_scores["token_res_ids"] = token_res_ids

        return cls(af3_scores)

    def __init__(self, af3_scores: dict):
        self.scores = af3_scores

    def to_file(self, file_path: Union[str, Path]):
        with open(file_path, "w") as f:
            json.dump(self.scores, f, indent=4)


def flatten(xss):
    return [x for xs in xss for x in xs]


def get_gap_indicies(*cif_objs) -> List[np.ndarray]:
    """
    Get the the gaps inbetween cif objects. Sometimes there is a discrepency
    between chain lengths between the modelling programs. This function is
    used to find where these discrepencies are.

    Args:
        *cif_objs: Multiple cif objects

    Returns:
        indicies: Dict with the chain_id as the key where the discrepency is located and
            the value is a list of indicies with -1 representing gaps

    """
    indicies: list = []

    if len(cif_objs) == 1:
        return indicies
    chain_lengths = [
        cif.chain_lengths(mode="residues", ligand_atoms=True) for cif in cif_objs
    ]

    assert all(
        [
            chain_lengths[0].keys() == chain_lengths[i].keys()
            for i in range(1, len(chain_lengths) - 1)
        ]
    )

    unequal_chain_lengths = [
        id_
        for id_ in chain_lengths[0].keys()
        if any(
            [
                chain_lengths[0][id_] != chain_lengths[i][id_]
                for (i, _) in enumerate(chain_lengths[1:], start=1)
            ]
        )
    ]

    for chain_id in chain_lengths[0]:
        if chain_id in unequal_chain_lengths:
            chain_atoms = [
                "".join([atom.element for atom in cif.get_atoms(chain_id=chain_id)])
                for cif in cif_objs
            ]

            longest = max(chain_atoms, key=len)

            for atom_str in chain_atoms:
                alignment = PairwiseAligner().align(longest, atom_str)
                indicies.append(alignment[0].indices[1])
        else:
            for _ in cif_objs:

                indicies.append(np.array([1] * chain_lengths[0][chain_id]))

    indicies = interleave_repeated(
        indicies, len(cif_objs), len(list(chain_lengths[0].keys()))
    )

    return indicies


def interleave_repeated(lst, n, chain_no):
    indicies = []
    chunks = [lst[i : i + n] for i in range(0, len(lst), n)]  # noqa: E203
    interleaved = [x for tup in zip_longest(*chunks) for x in tup if x is not None]

    for i in range(0, len(interleaved), chain_no):
        tmp_lst = []
        for j in range(chain_no):
            tmp_lst.extend(interleaved[i + j])
        indicies.append(tmp_lst)

    return indicies


def insert_none_by_minus_one(indices, values):
    result = []
    value_index = 0

    for idx in indices:
        if idx == -1:
            result.append(None)
        else:
            result.append(values[value_index])
            value_index += 1

    assert len(indices) == len(result)

    return result


def make_dummy_m8_file(run_json, output_dir):
    """
    Make a dummy m8 file with the templates from the run JSON file
    """
    with open(run_json) as f:
        input_json = json.load(f)

    templates = {}
    for sequence in input_json["sequences"]:
        if "protein" not in sequence:
            continue
        for id_ in sequence["protein"]["id"]:
            for template in sequence["protein"]["templates"]:
                if id_ in templates:
                    templates[id_].append(
                        template["mmcif"].split("\n")[0].split("_")[1]
                    )
                else:
                    templates[id_] = [template["mmcif"].split("\n")[0].split("_")[1]]

    m8_file = output_dir / "dummy.m8"
    if not templates:
        return None
    table = []

    for id_ in templates:
        for template in templates[id_]:
            table.append([id_, template, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    pd.DataFrame(table).to_csv(m8_file, sep="\t", header=False, index=False)

    return m8_file



from pathlib import Path
from typing import Union

from abcfold.output.file_handlers import CifFile, ConfidenceJsonFile
from abcfold.output.utils import Af3Pae


class AlphafoldOutput:
    def __init__(
        self,
        af3_output_dir: Union[str, Path],
        input_params: dict,
        name: str,
    ):
        """
        Object to process the output of an AlphaFold3 run

        Args:
            af3_output_dir (Union[str, Path]): Path to the AlphaFold3 output directory
            input_params (dict): Dictionary containing the input parameters used for the
            AlphaFold3 run
            name (str): Name given to the AlphaFold3 run

        Attributes:
            output_dir (Path): Path to the AlphaFold3 output directory
            input_params (dict): Dictionary containing the input parameters used for the
            AlphaFold3 run
            name (str): Name given to the AlphaFold3 run
            output (dict): Dictionary containing the processed output the contents
            of the AlphaFold3 output directory. The dictionary is structured as follows:

            {
                "seed-1": {
                    1: {
                        "cif": CifFile,
                        "json": ConfidenceJsonFile
                    },
                    2: {
                        "cif": CifFile,
                        "json": ConfidenceJsonFile
                    }
                },
                etc...
            }
            This is different to the boltz and chai equivalent as they do not have seeds
        """
        self.output_dir = Path(af3_output_dir)
        self.input_params = input_params

        if not self.output_dir.name.startswith("alphafold3"):
            self.output_dir = self.output_dir.rename(
                self.output_dir.parent.joinpath(f"alphafold3_{name}")
            )

        self.output = self.process_af3_output()
        self.seeds = list(self.output.keys())
        self.cif_files = {
            seed: [value["cif"] for value in self.output[seed].values()]
            for seed in self.seeds
        }
        self.af3_pae_files = {
            seed: [value["af3_pae"] for value in self.output[seed].values()]
            for seed in self.seeds
        }
        self.pae_to_af3()

        self.scores_files = {
            seed: [value["summary"] for value in self.output[seed].values()]
            for seed in self.seeds
        }
        self.input_json = list(self.output_dir.glob("*_data.json"))[0]

    def process_af3_output(self):
        """
        Process the output of an AlphaFold3 run

        """
        file_groups = {}
        for pathway in self.output_dir.iterdir():
            if pathway.is_dir():
                seed = pathway.name.split("_")[0]
                sample = pathway.name.split("-")[-1]
                if seed not in file_groups:
                    file_groups[seed] = {}
                if not sample.isdigit():
                    continue
                sample = int(sample)
                if sample not in file_groups[seed]:
                    file_groups[seed][sample] = {}
                for file in pathway.iterdir():
                    if file.suffix == ".cif":
                        cif_file = CifFile(str(file), self.input_params)
                        cif_file.name = f"Alphafold3_{seed}_{sample}"

                        cif_file = self.reorder_chains(cif_file)
                        cif_file.to_file(str(file))
                        file_groups[seed][sample]["cif"] = cif_file
                    elif file.suffix == ".json" and "summary" not in file.stem:
                        file_groups[seed][sample]["af3_pae"] = ConfidenceJsonFile(
                            str(file)
                        )
                    elif file.suffix == ".json" and "summary" in file.stem:
                        file_groups[seed][sample]["summary"] = ConfidenceJsonFile(
                            str(file)
                        )
                    else:
                        continue

        for seed in file_groups:
            file_groups[seed] = {
                sample: file_groups[seed][sample]
                for sample in sorted(file_groups[seed])
            }
        return file_groups

    def get_chain_ids(self) -> list:
        ids: list = []
        assert "sequences" in self.input_params
        for sequences in self.input_params["sequences"]:
            for sequence_type, sequence_information in sequences.items():
                if "id" not in sequence_information:
                    continue
                sequence_ids = sequence_information["id"]
                if isinstance(sequence_ids, list):
                    ids.extend(sequence_ids)
                elif isinstance(sequence_ids, str):
                    ids.append(sequence_ids)

        return ids

    def pae_to_af3(self):
        """
        Convert the PAE data from Boltz-1 to the format used by Alphafold3

        Returns:
            None
        """
        new_pae_files = {}
        for seed in self.seeds:
            for i, (pae_file, cif_file) in enumerate(
                zip(self.af3_pae_files[seed], self.cif_files[seed])
            ):
                pae = Af3Pae.from_alphafold3(
                    pae_file.data,
                    cif_file,
                )

                out_name = pae_file.pathway

                pae.to_file(out_name)

                if seed not in new_pae_files:
                    new_pae_files[seed] = []
                new_pae_files[seed].append(ConfidenceJsonFile(out_name))

        self.af3_pae_files = new_pae_files
        self.output = {
            seed: {
                i: {
                    "cif": cif_file,
                    "af3_pae": new_pae_files[seed][i],
                    "summary": self.output[seed][i]["summary"],
                }
                for i, cif_file in enumerate(self.cif_files[seed])
            }
            for seed in self.seeds
        }

    def reorder_chains(self, cif_file: CifFile) -> CifFile:
        """
        Function to update the chain order in the CIF file according to the ids in the
        input file

        Args:
            cif_file (CifFile): CifFile object to update the chain labels for

        """
        if [chain.id for chain in cif_file.get_chains()] == self.get_chain_ids():
            return cif_file
        cif_file.reorder_chains(self.get_chain_ids())
        return cif_file



import json
import tempfile
from pathlib import Path

from abcfold.boltz1.af3_to_boltz1 import DELIM, BoltzYaml


def test_af3_to_boltz(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        boltz_yaml = BoltzYaml(temp_dir)

        yaml_string_multi = boltz_yaml.json_to_yaml(test_data.test_inputAB_json)

        yaml_string_multi = yaml_string_multi.split("\n")

        assert yaml_string_multi[0] == "version: 1"
        assert yaml_string_multi[1] == "sequences:"
        assert yaml_string_multi[2] == f"{DELIM}- protein:"
        assert yaml_string_multi[3] == f"{DELIM}{DELIM}id: [A, B]"
        assert yaml_string_multi[4] == f"{DELIM}{DELIM}sequence: GMRES"
        assert yaml_string_multi[5] == f"{DELIM}- protein:"
        assert yaml_string_multi[6] == f"{DELIM}{DELIM}id: C"
        assert yaml_string_multi[7] == f"{DELIM}{DELIM}sequence: YANEN"
        assert yaml_string_multi[8] == f"{DELIM}- ligand:"
        assert yaml_string_multi[9] == f"{DELIM}{DELIM}id: [D, E]"
        assert yaml_string_multi[10] == f"{DELIM}{DELIM}ccd: ATP"
        assert yaml_string_multi[11] == f"{DELIM}- ligand:"
        assert yaml_string_multi[12] == f"{DELIM}{DELIM}id: F"
        assert yaml_string_multi[13] == f"{DELIM}{DELIM}smiles: CC(=O)OC1C[NH+]2CCC1CC2"

        yaml_string_bonds = boltz_yaml.json_to_yaml(test_data.test_inputA_json)
        yaml_string_bonds = yaml_string_bonds.split("\n")

        print(yaml_string_bonds)
        assert yaml_string_bonds[0] == "version: 1"
        assert yaml_string_bonds[1] == "sequences:"
        assert yaml_string_bonds[2] == f"{DELIM}- protein:"
        assert yaml_string_bonds[3] == f"{DELIM}{DELIM}id: A"
        assert (
            yaml_string_bonds[4]
            == f"{DELIM}{DELIM}sequence: GMRESYANENQFGFKTINSDIHKIVIVGGYGKLGGLFARY\
LRASGYPISILDREDWAVAESILANADVVIVSVPINLTLETIERLKPYLTENMLLADLTSVKREPLAKMLEVHTGAVLGLHPMFG\
ADIASMAKQVVVRCDGRFPERYEWLLEQIQIWGAKIYQTNATEHDHNMTYIQALRHFSTFANGLHLSKQPINLANLLALSSPIYRL\
ELAMIGRLFAQDAELYADIIMDKSENLAVIETLKQTYDEALTFFENNDRQGFIDAFHKVRDWFGDYSEQFLKESRQLLQQANDLKQG"
        )

        with open(test_data.test_inputAB_json, "r") as f:
            json_dict = json.load(f)

        yaml_string = boltz_yaml.json_to_yaml(json_dict)
        yaml_string = yaml_string.split("\n")

        assert yaml_string[0] == "version: 1"
        assert yaml_string[1] == "sequences:"
        assert yaml_string_multi[2] == f"{DELIM}- protein:"
        assert yaml_string_multi[3] == f"{DELIM}{DELIM}id: [A, B]"
        assert yaml_string_multi[4] == f"{DELIM}{DELIM}sequence: GMRES"
        assert yaml_string_multi[5] == f"{DELIM}- protein:"
        assert yaml_string_multi[6] == f"{DELIM}{DELIM}id: C"
        assert yaml_string_multi[7] == f"{DELIM}{DELIM}sequence: YANEN"
        assert yaml_string_multi[8] == f"{DELIM}- ligand:"
        assert yaml_string_multi[9] == f"{DELIM}{DELIM}id: [D, E]"
        assert yaml_string_multi[10] == f"{DELIM}{DELIM}ccd: ATP"
        assert yaml_string_multi[11] == f"{DELIM}- ligand:"
        assert yaml_string_multi[12] == f"{DELIM}{DELIM}id: F"
        assert yaml_string_multi[13] == f"{DELIM}{DELIM}smiles: CC(=O)OC1C[NH+]2CCC1CC2"


def test_boltz_output_yaml(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        boltz_yaml = BoltzYaml(temp_dir)

        assert not boltz_yaml.yaml_string
        boltz_yaml.json_to_yaml(test_data.test_inputAB_json)

        assert boltz_yaml.yaml_string

        file_name = Path(temp_dir) / "output.yaml"
        boltz_yaml.write_yaml(file_name)

        assert file_name.exists()
        assert file_name.is_file()


def test_boltz_output_msa(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        boltz_yaml = BoltzYaml(temp_dir)

        assert boltz_yaml.msa_file == "null"

        boltz_yaml.json_to_yaml(test_data.test_inputAmsa_json)

        assert boltz_yaml.msa_file != "null"
        assert boltz_yaml.msa_file.exists()
        assert boltz_yaml.msa_file.is_file()


def test_af3_data_json_to_yaml(output_objs):
    try:
        af3_json = output_objs.af3_output.input_json
        with tempfile.TemporaryDirectory() as temp_dir:
            boltz_yaml = BoltzYaml(temp_dir)
            boltz_yaml.json_to_yaml(af3_json)
    except TypeError:
        assert False


def test_extra_ligand_ids():
    difficult_params = {
        "name": "Nightmare",
        "modelSeeds": [42],
        "sequences": [
            {
                "protein": {
                    "id": "A",
                    "sequence": "PVLSCGEWQL",
                    "modifications": [
                        {"ptmType": "HY3", "ptmPosition": 1},
                        {"ptmType": "P1L", "ptmPosition": 5},
                    ],
                }
            },
            {"protein": {"id": "B", "sequence": "QIQLVQSGPELKKPGET"}},
            {"protein": {"id": "C", "sequence": "DVLMIQTPLSLPVS"}},
            {"ligand": {"id": ["F"], "ccdCodes": ["ATP"]}},
            {"ligand": {"id": "I", "ccdCodes": ["NAG", "FUC", "FUC"]}},
            {"dna": {"id": ["D", "K"], "sequence": "AGCT"}},
            {"rna": {"id": "L", "sequence": "AGCU"}},
            {"ligand": {"id": "Z", "smiles": "CC(=O)OC1C[NH+]2CCC1CC2"}},
        ],
        "bondedAtomPairs": [
            [["A", 1, "CA"], ["B", 1, "CA"]],
            [["C", 7, "CA"], ["A", 10, "CA"]],
            [["I", 1, "O3"], ["I", 2, "C1"]],
            [["I", 2, "C1"], ["I", 3, "C1"]],
        ],
        "dialect": "scouse",
        "version": 2,
    }

    with tempfile.TemporaryDirectory() as temp_dir:
        boltz_yaml = BoltzYaml(temp_dir)
        yaml_string = boltz_yaml.json_to_yaml(difficult_params)
        yaml_string = yaml_string.split("\n")

        assert yaml_string[0] == "version: 1"
        assert yaml_string[1] == "sequences:"
        assert yaml_string[2] == f"{DELIM}- protein:"
        assert yaml_string[3] == f"{DELIM}{DELIM}id: A"
        assert yaml_string[4] == f"{DELIM}{DELIM}sequence: PVLSCGEWQL"
        assert yaml_string[5] == f"{DELIM}{DELIM}modifications:"
        assert yaml_string[6] == f"{DELIM}{DELIM}{DELIM}- position: 1"
        assert yaml_string[7] == f"{DELIM}{DELIM}{DELIM}  ccd: HY3"
        assert yaml_string[8] == f"{DELIM}{DELIM}{DELIM}- position: 5"
        assert yaml_string[9] == f"{DELIM}{DELIM}{DELIM}  ccd: P1L"
        assert yaml_string[10] == f"{DELIM}- protein:"
        assert yaml_string[11] == f"{DELIM}{DELIM}id: B"
        assert yaml_string[12] == f"{DELIM}{DELIM}sequence: QIQLVQSGPELKKPGET"
        assert yaml_string[13] == f"{DELIM}- protein:"
        assert yaml_string[14] == f"{DELIM}{DELIM}id: C"
        assert yaml_string[15] == f"{DELIM}{DELIM}sequence: DVLMIQTPLSLPVS"
        assert yaml_string[16] == f"{DELIM}- ligand:"
        assert yaml_string[17] == f"{DELIM}{DELIM}id: [F]"
        assert yaml_string[18] == f"{DELIM}{DELIM}ccd: ATP"
        assert yaml_string[19] == f"{DELIM}- ligand:"
        assert yaml_string[20] == f"{DELIM}{DELIM}id: I"
        assert yaml_string[21] == f"{DELIM}{DELIM}ccd: NAG"
        assert yaml_string[22] == f"{DELIM}- ligand:"
        assert yaml_string[23] == f"{DELIM}{DELIM}id: E"
        assert yaml_string[24] == f"{DELIM}{DELIM}ccd: FUC"
        assert yaml_string[25] == f"{DELIM}- ligand:"
        assert yaml_string[26] == f"{DELIM}{DELIM}id: G"
        assert yaml_string[27] == f"{DELIM}{DELIM}ccd: FUC"
        assert yaml_string[28] == f"{DELIM}- dna:"
        assert yaml_string[29] == f"{DELIM}{DELIM}id: [D, K]"
        assert yaml_string[30] == f"{DELIM}{DELIM}sequence: AGCT"
        assert yaml_string[31] == f"{DELIM}- rna:"
        assert yaml_string[32] == f"{DELIM}{DELIM}id: L"
        assert yaml_string[33] == f"{DELIM}{DELIM}sequence: AGCU"
        assert yaml_string[34] == f"{DELIM}- ligand:"
        assert yaml_string[35] == f"{DELIM}{DELIM}id: Z"
        assert yaml_string[36] == f"{DELIM}{DELIM}smiles: CC(=O)OC1C[NH+]2CCC1CC2"
        assert yaml_string[37] == "constraints:"
        assert yaml_string[38] == f"{DELIM}- bond:"
        assert yaml_string[39] == f"{DELIM}{DELIM}atom1: ['A', 1, 'CA']"
        assert yaml_string[40] == f"{DELIM}{DELIM}atom2: ['B', 1, 'CA']"
        assert yaml_string[41] == f"{DELIM}- bond:"
        assert yaml_string[42] == f"{DELIM}{DELIM}atom1: ['C', 7, 'CA']"
        assert yaml_string[43] == f"{DELIM}{DELIM}atom2: ['A', 10, 'CA']"
        assert yaml_string[44] == f"{DELIM}- bond:"
        assert yaml_string[45] == f"{DELIM}{DELIM}atom1: ['I', 1, 'O3']"
        assert yaml_string[46] == f"{DELIM}{DELIM}atom2: ['E', 1, 'C1']"
        assert yaml_string[47] == f"{DELIM}- bond:"
        assert yaml_string[48] == f"{DELIM}{DELIM}atom1: ['E', 1, 'C1']"
        assert yaml_string[49] == f"{DELIM}{DELIM}atom2: ['G', 1, 'C1']"



import argparse
import configparser
import os
import tempfile
from pathlib import Path

import pytest

from abcfold.abcfold import run


@pytest.mark.skipif(
    os.getenv("FULL_TEST") != "true",
    reason="This test is skipped in the full test suite",
)
def test_abcfold(test_data):
    with tempfile.TemporaryDirectory() as temp_dir_str:
        temp_dir = Path(temp_dir_str)
        config_file_path = temp_dir / "config.ini"

        config_file_path.write_text(
            """
[Databases]
model_params = None
database_dir = None
"""
        )

        defaults = {}
        config = configparser.SafeConfigParser()

        if config_file_path.exists():
            config.read(str(config_file_path))
            defaults.update(dict(config.items("Databases")))

        args = argparse.Namespace(
            input_json=test_data.test_inputAB_json,
            output_dir=temp_dir / "output",
            model_params=str(temp_dir),
            database_dir=str(temp_dir),
            # interactive=False,
            override=True,
            alphafold3=False,
            boltz1=True,
            chai1=True,
            mmseqs2=True,
            output_json=None,
            templates=False,
            num_templates=0,
            custom_template=False,
            custom_template_chain=None,
            target_id=None,
            save_input=False,
        )

        run(
            args,
            config_file=config_file_path,
            defaults=defaults,
            config=config,
        )

        out_dirs = [x.name for x in Path(args.output_dir).iterdir()]
        print(out_dirs)

        assert "chai1" in out_dirs
        assert "boltz-1_inputAB_mmseqs" in out_dirs



# Run the code for the PAE plots

import logging
import shutil
import subprocess
from multiprocessing import Process
from pathlib import Path
from typing import Dict, List, Union

import pandas as pd

from abcfold.output.alphafold3 import AlphafoldOutput
from abcfold.output.boltz import BoltzOutput
from abcfold.output.chai import ChaiOutput
from abcfold.output.file_handlers import CifFile

logger = logging.getLogger(__name__)

CSSPATHS = {
    "A": "paeViewerStandaloneLayoutAF3.css",
    "B": "paeViewerStandaloneLayoutBoltz.css",
    "C": "paeViewerStandaloneLayoutChai.css",
}

PAEVIEWER = Path(__file__).parent.joinpath(
    "pae-viewer-main", "standalone", "pae_viewer.py"
)

CREATETEMPLATE = Path(__file__).parent.joinpath(
    "pae-viewer-main", "standalone", "create_template.py"
)

INCLUDE_EXTS = [
    ".css",
    ".js",
    ".html",
    ".tpl",
    ".svg",
    ".ico",
    ".png",
]


COLUMNS = [
    "Protein1",
    "SeqPos1",
    "Protein2",
    "SeqPos2",
    "RestraintSatisfied",
]


def create_pae_plots(
    outputs: list,
    output_dir: Union[str, Path],
) -> Dict[str, str]:
    """
    Create PAE html plots for the outputs

    Args:
        outputs: List of outputs to create plots for
        output_dir: Output directory for the plots

    Returns:
        Pathway plot dictionary with the key as the plot path and value as the plot path

    Raises:
        ValueError: If the output type is invalid


    """
    pathway_plot: Dict[str, str] = {}
    run_scripts: List[list] = []
    template_files: List[Path] = []
    output_dir = Path(output_dir)

    # start by copying the pae viewer files to the output directory
    copy_pae_viewer_files(output_dir.joinpath(".pae_viewer"))

    for output in outputs:
        plots_dir = (
            Path(output_dir)
            if output_dir
            else output.output_dir.parent.joinpath(".plots")
        )
        plots_dir.mkdir(exist_ok=True)

        if isinstance(output, BoltzOutput):
            css_path = CSSPATHS["B"]
            template_file = plots_dir.joinpath("boltz_template.html")
            template_files.append(template_file)
            cmd = get_template_run_script(
                "ABCFold - Boltz-1 Output",
                css_path,
                template_file,
                output_dir.joinpath(".pae_viewer"),
            )
            run_script(cmd)

        elif isinstance(output, ChaiOutput):
            css_path = CSSPATHS["C"]
            template_file = plots_dir.joinpath("chai_template.html")
            template_files.append(template_file)
            cmd = get_template_run_script(
                "ABCFold - Chai-1 Output",
                css_path,
                template_file,
                output_dir.joinpath(".pae_viewer"),
            )
            run_script(cmd)

        elif isinstance(output, AlphafoldOutput):
            css_path = CSSPATHS["A"]
            template_file = plots_dir.joinpath("af3_template.html")
            template_files.append(template_file)
            cmd = get_template_run_script(
                "ABCFold - AlphaFold-3 Output",
                css_path,
                template_file,
                output_dir.joinpath(".pae_viewer"),
            )
            run_script(cmd)

            for seed in output.seeds:
                run_scripts.extend(
                    prepare_scripts(
                        output.cif_files[seed],
                        output.af3_pae_files[seed],
                        plots_dir,
                        pathway_plot,
                        template_file,
                        True,
                    )
                )

            continue
        else:
            logger.error("Invalid output type")
            raise ValueError()

        run_scripts.extend(
            prepare_scripts(
                output.cif_files,
                output.af3_pae_files,
                plots_dir,
                pathway_plot,
                template_file,
                False,
            )
        )

    processes = [Process(target=run_script, args=(script,)) for script in run_scripts]
    for process in processes:
        process.start()
    for process in processes:
        process.join()

    for template_file in template_files:
        template_file.unlink()

    for csv_file in output_dir.glob("*.csv"):
        csv_file.unlink()

    return pathway_plot


def prepare_scripts(
    cif_files, pae_files, plots_dir, pathway_plot, template_file, is_af3=False
):

    scripts = []
    for cif_file, pae_file in zip(cif_files, pae_files):
        name_stem = f"{pae_file.pathway.stem}\
{'_' + pae_file.pathway.parent.stem if is_af3 else ''}_{'af3_' if is_af3 else ''}"

        clashes_csv_file = plots_dir.joinpath(f"{name_stem}clashes.csv")
        clashes_csv(cif_file, clashes_csv_file)

        labels = [f"Chain-{chain}" for chain in cif_file.chain_lengths()]
        plot_pathway = plots_dir.joinpath(f"{name_stem}pae_plot.html")
        pae_viewer_script = get_pae_run_script(
            cif_file.pathway,
            labels,
            pae_file.pathway,
            plot_pathway,
            template_file,
            clashes_csv_file,
        )
        pathway_plot[str(cif_file.pathway)] = str(plot_pathway)
        scripts.append(pae_viewer_script)
    return scripts


def run_script(cmd):
    with subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as proc:
        stdout, stderr = proc.communicate()
        proc.wait()

        if proc.returncode != 0:
            raise subprocess.CalledProcessError(proc.returncode, cmd, stderr)


def get_pae_run_script(
    cif_path: Union[str, Path],
    labels: List[str],
    pae_path: Union[str, Path],
    output_file: Union[str, Path],
    template_file: Union[str, Path],
    clashes_csv_file: Union[str, Path],
):
    """
    Get the command to run the PAE viewer script

    Args:
        cif_path: Path to the CIF file
        labels: List of labels for the chains
        pae_path: Path to the PAE file
        output_file: Path to the output file
        template_file: Path to the template file
        clashes_csv: Path to the clashes CSV file
    Returns:
        Command to run the PAE viewer script
    """

    cif_path = Path(cif_path)
    pae_path = Path(pae_path)
    output_file = Path(output_file)
    template_file = Path(template_file)
    clashes_csv_file = Path(clashes_csv_file)

    cmd = []
    labels_string = ";".join(labels)
    # labels_string = rf'"{labels_string}"'

    cmd.append("python")
    cmd.append(str(PAEVIEWER.resolve()))
    cmd.append("--structure")
    cmd.append(str(cif_path.resolve()))
    cmd.append("--labels")

    cmd.append(labels_string)
    cmd.append("--scores")
    cmd.append(str(pae_path.resolve()))
    cmd.append("--output_file")
    cmd.append(str(output_file.resolve()))
    cmd.append("--template_file")
    cmd.append(str(template_file.resolve()))
    cmd.append("--crosslinks")
    cmd.append(str(clashes_csv_file.resolve()))
    return cmd


def get_template_run_script(
    title: str,
    standalonecss: str,
    output_file: Union[str, Path],
    src_path: Union[str, Path],
):
    """ """
    output_file = Path(output_file)
    src_path = Path(src_path)
    cmd = []
    cmd.append("python")
    cmd.append(str(CREATETEMPLATE.resolve()))
    cmd.append("--title")
    # title = rf'"{title}"'
    cmd.append(title)
    cmd.append("--standalonecss")
    cmd.append(standalonecss)
    cmd.append("--output_file")
    cmd.append(str(output_file.resolve()))
    cmd.append("--src_path")
    cmd.append(str(src_path.resolve()))

    return cmd


def make_dir(output_dir: Union[str, Path]):
    output_dir = Path(output_dir)
    if output_dir.exists():
        return output_dir
    output_dir.mkdir(parents=True)
    return output_dir


def copy_pae_viewer_files(output_dir: Union[str, Path]):
    output_dir = Path(output_dir)
    if not output_dir.exists():
        output_dir.mkdir(parents=True)
    pae_viewer_main_pathway = Path(__file__).parent.joinpath("pae-viewer-main")
    pae_viewer_files = pae_viewer_main_pathway.rglob("*")

    for file_ in pae_viewer_files:
        if file_.suffix in INCLUDE_EXTS:
            # print out the filepathway from the current directory
            subdirs = list(file_.relative_to(pae_viewer_main_pathway).parts)
            create_subdirs(output_dir, subdirs)
            # copy the file to the output directory
            new_file = output_dir.joinpath(*subdirs)
            if not new_file.exists():
                shutil.copy2(file_, new_file)


def create_subdirs(output_dir: Union[str, Path], subdirs: List[str]):
    output_dir = Path(output_dir)
    subdirs = subdirs[:-1]
    new_subdirs = output_dir
    for subdir in subdirs:
        new_subdirs = new_subdirs.joinpath(subdir)
        if not new_subdirs.exists():
            new_subdirs.mkdir()


def clashes_csv(cif_file: CifFile, output_name: Union[str, Path]):
    output_name = Path(output_name)

    _, clashes = cif_file.check_clashes()
    df = pd.DataFrame(columns=COLUMNS)

    for clash in clashes:
        atom1, atom2 = clash
        chain_id1, chain_id2 = atom1.get_full_id()[2], atom2.get_full_id()[2]
        res1, res2 = atom1.get_full_id()[3][1], atom2.get_full_id()[3][1]
        df.loc[len(df)] = [
            f"Chain-{chain_id1}",
            res1,
            f"Chain-{chain_id2}",
            res2,
            False,
        ]

    df.to_csv(output_name, index=False)



def test_process_af3_output(test_data, output_objs):
    af3_output = output_objs.af3_output
    assert str(af3_output.output_dir) == test_data.test_alphafold3_6BJ9_

    assert "seed-1" in af3_output.output

    files = af3_output.output["seed-1"]

    assert 0 in files
    assert 1 in files

    assert "cif" in files[0]
    assert "af3_pae" in files[0]

    assert "cif" in files[1]
    assert "af3_pae" in files[1]

    json_file = af3_output.af3_pae_files["seed-1"][0]

    assert "atom_plddts" in json_file.data
    assert "pae" in json_file.data



import json
import logging
import os
import shutil
import sys
import time
from io import StringIO
from pathlib import Path
from typing import Mapping, Optional, Union

from Bio import Align
from Bio.PDB import MMCIFIO, MMCIFParser
from colorama import Fore, Style

logger = logging.getLogger("logger")


# Custom formatter for colored logging
class ColoredFormatter(logging.Formatter):
    # Define color codes for each log level
    LEVEL_COLORS = {
        logging.DEBUG: Fore.BLUE,
        logging.INFO: Fore.WHITE,
        logging.WARNING: Fore.YELLOW,
        logging.ERROR: Fore.RED,
        logging.CRITICAL: Fore.RED + Style.BRIGHT,
    }

    def format(self, record):
        # Get the color for the log level
        level_color = self.LEVEL_COLORS.get(record.levelno, "")
        # Format the log message
        formatted_message = super().format(record)
        # Return the message with the color added
        return f"{level_color}{formatted_message}{Style.RESET_ALL}"


# Set up logging
def setup_logger():
    logger = logging.getLogger("logger")
    logger.setLevel(logging.DEBUG)  # Set the minimum logging level

    # Create a stream handler (output to console)
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)

    # Set the custom formatter
    formatter = ColoredFormatter("%(asctime)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)

    # Add the handler to the logger
    logger.addHandler(handler)

    return logger


def get_chains(mmcif_file):
    """Return a list of chains in a MMCIF file."""
    parser = MMCIFParser(QUIET=True)
    structure = parser.get_structure("template", mmcif_file)
    chains = []
    for model in structure:
        for chain in model:
            chains.append(chain.id)
    return chains


def extract_sequence_from_mmcif(mmcif_file):
    """Extract the sequence from a MMCIF file."""
    parser = MMCIFParser(QUIET=True)
    structure = parser.get_structure("template", mmcif_file)
    sequence = ""
    model = structure[0]  # Assuming one model/chain only
    for chain in model:
        for residue in chain:
            if residue.id[0] == " ":  # Exclude heteroatoms
                sequence += residue.resname[0]  # Simplified to take the first letter
    return sequence


# Code from https://github.com/google-deepmind/alphafold3
def query_to_hit_mapping(
    query_aligned: str, template_aligned: str
) -> Mapping[int, int]:
    """0-based query index to hit index mapping."""
    query_to_hit_mapping_out = {}
    hit_index = 0
    query_index = 0
    for q_char, t_char in zip(query_aligned, template_aligned):
        # Gap inserted in the template
        if q_char == "-":
            query_index += 1
        # Deleted residue in the template (would be a gap in the query).
        elif t_char == "-":
            hit_index += 1
        # Normal aligned residue, in both query and template. Add to mapping.
        else:
            query_to_hit_mapping_out[query_index] = hit_index
            query_index += 1
            hit_index += 1
    return query_to_hit_mapping_out


def align_and_map(query_seq, template_seq):
    """Align two sequences and map the indices."""
    # Perform pairwise alignment
    aligner = Align.PairwiseAligner()
    alignments = aligner.align(query_seq, template_seq)
    alignment = alignments[0]  # Take the best alignment

    formatted_alignment = alignment._format_generalized().replace(" ", "")
    query_aligned, _, template_aligned, _ = formatted_alignment.split("\n")

    # Map the aligned sequences
    aligned_mapping = query_to_hit_mapping(query_aligned, template_aligned)

    query_indices = []
    template_indices = []
    for template_index, query_index in aligned_mapping.items():
        query_indices.append(query_index)
        template_indices.append(template_index)

    return query_indices, template_indices


def get_mmcif(
    cif,
    pdb_id,
    chain_id,
    start,
    end,
    tmpdir=None,
):
    """
    Extract a chain from a CIF file and return a new CIF string with only the
    specified chain, residues and metadata.
    """

    parser = MMCIFParser(QUIET=True)
    structure = parser.get_structure(pdb_id, cif)

    # Extract release date from the CIF file
    mmcif_dict = parser._mmcif_dict
    headers_to_keep = [
        "_entry.id",
        "_entry.title",
        "_entry.deposition_date",
        "_pdbx_audit_revision_history.revision_date",
    ]
    filtered_metadata = {
        key: mmcif_dict[key] for key in headers_to_keep if key in mmcif_dict
    }

    # Make metadata if missing
    if "_pdbx_audit_revision_history.revision_date" not in filtered_metadata:
        filtered_metadata["_pdbx_audit_revision_history.revision_date"] = time.strftime(
            "%Y-%m-%d"
        )

    # For multimodel templates (e.g. NMR) pick a single representative model
    # add a copy of the first model to the structure

    if len(structure) > 1:
        for model_index in range(1, len(structure)):
            structure.detach_child(structure[model_index].get_id())

    for model in structure:
        chain_to_del = []
        for chain in model:
            if chain.id != chain_id:
                chain_to_del.append(chain.id)
                continue

        for unwanted_chain in chain_to_del:
            model.detach_child(unwanted_chain)

        for chain in model:
            res_to_del = []
            for i, res in enumerate(chain):
                rel_pos = i + 1
                if rel_pos < start or rel_pos > end or res.id[0] != " ":
                    res_to_del.append(res)

            for res in res_to_del:
                chain.detach_child(res.id)

    # Save the filtered structure to a new CIF file
    io = MMCIFIO()
    io.set_structure(structure)
    filtered_output = (
        f"{pdb_id}_{chain_id}.cif"
        if tmpdir is None
        else f"{tmpdir}/{pdb_id}_{chain_id}.cif"
    )
    io.save(filtered_output)

    # Parse the filtered structure to get the modified MMCIF with no metadata
    structure = parser.get_structure(pdb_id, filtered_output)
    mmcif_dict = parser._mmcif_dict

    # Add the filtered metadata to the MMCIF dictionary
    mmcif_dict.update(filtered_metadata)

    # Save the modified MMCIF with wanted metadata to a string
    string_io = StringIO()
    io.set_dict(mmcif_dict)
    io.save(string_io)

    os.unlink(filtered_output)

    return string_io.getvalue()


# runs for each sequence in the input json
def get_custom_template(
    sequence,
    target_id,
    custom_template,
    custom_template_chain,
):
    """Add a custom template to the input json"""

    # add code here to run the custom template but move the res to output
    # Keep existing templates if they're present
    if "templates" not in sequence["protein"]:
        templates = []
    else:
        templates = sequence["protein"]["templates"]

    input_sequence = sequence["protein"]["sequence"]
    seq_id = sequence["protein"]["id"]
    if isinstance(seq_id, list):
        if target_id and target_id not in seq_id:
            return sequence
    if isinstance(seq_id, str):
        if target_id and target_id != seq_id:
            return sequence

    if not os.path.exists(custom_template):
        msg = f"Custom template file {custom_template} not found"
        logger.critical(msg)
        raise FileNotFoundError()

    chain_info = get_chains(custom_template)
    if len(chain_info) != 1 and not custom_template_chain:
        msg = f"Custom template file {custom_template} contains \
{len(chain_info)} chains. Please specify the chain to use with --custom_template_chain"
        raise ValueError(msg)

    if custom_template_chain and custom_template_chain not in chain_info:
        msg = f"Custom template file {custom_template} does not \
contain chain {custom_template_chain}"
        raise ValueError(msg)

    if not custom_template_chain:
        custom_template_chain = chain_info[0]

    template = {}
    cif_str = get_mmcif(
        custom_template,
        "custom",
        custom_template_chain,
        1,
        len(input_sequence),
    )

    template["mmcif"] = cif_str
    template_seq = extract_sequence_from_mmcif(StringIO(cif_str))
    query_indices, template_indices = align_and_map(input_sequence, template_seq)

    template["queryIndices"] = query_indices
    template["templateIndices"] = template_indices

    # Add the custom template to the start of the templates list
    templates.insert(0, template)

    # Add template to the json
    sequence["protein"]["templates"] = templates

    # Save the output json
    return sequence


def make_dir(dir_path: Union[str, Path], overwrite: bool = False):
    """
    Make a directory and return the Path object.

    Args:
        dir_path: The path to the directory to create.
        overwrite: Whether to delete the directory if it already exists.

    Returns:
        The Path object for the created directory.
    """
    dir_path = Path(dir_path)
    if dir_path.exists():
        if overwrite:
            shutil.rmtree(dir_path)
        else:
            logger.error(
                f"Directory {dir_path} already exists, use --override to replace it"
            )
            raise FileExistsError()

    dir_path.mkdir(parents=True, exist_ok=True)
    return dir_path


def check_input_json(
    input_json: Union[str, Path],
    output_dir: Optional[Union[str, Path]] = None,
    use_af3_templates: bool = False,
    test: bool = False,
):
    """
    Check the input json file for missing fields and add default values.

    Args:
        input_json: The path to the input json file.
        use_af3_templates: Whether to use the AlphaFold3 templates.

    Returns:
        None
    """
    input_json = Path(input_json)

    output_json = (
        input_json.parent.joinpath("abc_" + input_json.name)
        if output_dir is None
        else Path(output_dir).joinpath("abc_" + input_json.name)
    )
    with open(input_json, "r") as f:
        input_data = json.load(f)

    for sequence in input_data["sequences"]:
        for sequence_type in sequence:
            if "unpairedMsaPath" in sequence[sequence_type]:
                msa_path = sequence[sequence_type]["unpairedMsaPath"]
                if not os.path.exists(msa_path):
                    logger.error(f"MSA file {msa_path} not found")
                    sys.exit(1)
                with open(msa_path, "r") as f:
                    msa = f.read()
                sequence[sequence_type]["unpairedMsa"] = msa
                del sequence[sequence_type]["unpairedMsaPath"]
            if "unpairedMsa" in sequence[sequence_type]:
                if "templates" not in sequence[sequence_type]:

                    sequence[sequence_type]["templates"] = (
                        None if use_af3_templates else []
                    )

                if "pairedMsa" not in sequence[sequence_type]:
                    sequence[sequence_type]["pairedMsa"] = ""

    if test:
        return input_data
    with open(output_json, "w") as f:
        json.dump(input_data, f, indent=4)

    return output_json


def make_dummy_af3_db(output_dir):
    dummy_af3_db = output_dir.joinpath("af3_db")
    dummy_files = [
        "bfd-first_non_consensus_sequences.fasta",
        "rfam_14_9_clust_seq_id_90_cov_80_rep_seq.fasta",
        "uniprot_all_2021_04.fa",
        "mgy_clusters_2022_05.fa",
        "nt_rna_2023_02_23_clust_seq_id_90_cov_80_rep_seq.fasta",
        "pdb_seqres_2022_09_28.fasta",
        "rnacentral_active_seq_id_90_cov_80_linclust.fasta",
        "uniref90_2022_05.fa"
        ]
    dummy_dirs = ["mmcif_files"]

    dummy_af3_db.mkdir(parents=True, exist_ok=True)
    for dummy_file in dummy_files:
        with open(dummy_af3_db.joinpath(dummy_file), "w") as f:
            f.write("")
    for dummy_dir in dummy_dirs:
        dummy_af3_db.joinpath(dummy_dir).mkdir(parents=True, exist_ok=True)

    return dummy_af3_db



import logging
import sys
from pathlib import Path

logger = logging.getLogger("logger")


def main_argpase_util(parser):
    parser.add_argument("input_json", help="Input sequence file")
    parser.add_argument("output_dir", help="Output directory")

    return parser


def mmseqs2_argparse_util(parser):
    parser.add_argument(
        "--mmseqs2",
        action="store_true",
        help="Use MMseqs2 for MSA",
    )
    parser.add_argument(
        "--templates", action="store_true", help="Include templates in the output json"
    )
    parser.add_argument(
        "--num_templates",
        type=int,
        default=20,
        help="Number of templates to include in the output json",
    )

    return parser


def custom_template_argpase_util(parser):
    parser.add_argument(
        "--target_id", nargs="+", help="Target id relating to the custom template"
    )
    parser.add_argument(
        "--custom_template",
        nargs="+",
        help="Custom template to include in the output json",
    )
    parser.add_argument(
        "--custom_template_chain",
        nargs="+",
        help="Custom template chain to include in the output json",
    )

    return parser


def prediction_argparse_util(parser):
    parser.add_argument(
        "--number_of_models",
        type=int,
        default=5,
        help="Number of models to generate",
    )
    parser.add_argument(
        "--num_recycles",
        type=int,
        default=10,
        help="Number of recycles to use during the inference",
    )
    return parser


def boltz_argparse_util(parser):
    parser.add_argument(
        "-b",
        "--boltz1",
        action="store_true",
        help="Run Boltz1",
    )
    if "--save_input" not in parser._option_string_actions:
        parser.add_argument(
            "--save_input",
            action="store_true",
            help="Save the input json file",
            default=False,
        )

    return parser


def chai_argparse_util(parser):
    parser.add_argument(
        "-c",
        "--chai1",
        action="store_true",
        help="Run Chai-1",
    )
    return parser


def alphafold_argparse_util(parser):

    parser.add_argument("--output_json", help="Output json file")
    # make the vartible saved as database_dir
    parser.add_argument(
        "--database",
        help="The Database directory for the generation of the MSA.",
        dest="database_dir",
        default=None,
    )

    parser.add_argument(
        "--model_params",
        help="The directory containing the model parameters",
        default=None,
    )

    parser.add_argument(
        "-a",
        "--alphafold3",
        action="store_true",
        help="Run Alphafold3",
    )

    parser.add_argument(
        "--override",
        help="Override the existing output directory, if it exists",
        action="store_true",
    )

    parser.add_argument(
        "--use_af3_template_search",
        action="store_true",
        help="If providing your own custom MSA or the you've ran `--mmseqs`, allow \
Alphafold3 to search for templates",
    )

    return parser


def visuals_argparse_util(parser):
    parser.add_argument(
        "--no_visuals",
        action="store_true",
        help="Do not generate the output pages, best for running on a cluster\
 without a display",
    )
    parser.add_argument(
        "--no_server",
        action="store_true",
        help="Do not start a local server to view the results, the output page is \
stil generated and is accessible in the output directory",
    )
    return parser


def raise_argument_errors(args):
    if not args.alphafold3 and not args.boltz1 and not args.chai1:
        logger.info(
            "Neither AlphaFold3, Boltz-1, or Chai-1 selected. Running AlphaFold3 \
by default"
        )
        args.alphafold3 = True

    if (
        args.alphafold3
        and (not args.database_dir or not Path(args.database_dir).exists())
        and not args.mmseqs2
    ):
        logger.error(f"Database directory not found: {args.database_dir}")
        sys.exit(1)
    elif (
        args.alphafold3
        and (not args.model_params or not Path(args.model_params).exists())
        and not args.mmseqs2
    ):
        logger.error(f"Model parameters directory not found: {args.model_params}")
        sys.exit(1)

    if args.templates and not args.mmseqs2:
        logger.error("Cannot include templates without using MMseqs2")
        sys.exit(1)

    if args.custom_template_chain and not args.custom_template:
        logger.error("Custom template chain provided without a custom template")
        sys.exit(1)

    if args.use_af3_template_search and not args.alphafold3:
        logger.error(
            "Cannot use the Alphafold3 template search without running Alphafold3"
        )
        sys.exit(1)

    if args.num_templates < 1:
        logger.error("Number of templates must be greater than 0")
        sys.exit(1)

    if args.num_recycles < 1:
        logger.error("Number of recycles must be greater than 0")
        sys.exit(1)

    if args.number_of_models < 1:
        logger.error("Number of models must be greater than 0")
        sys.exit(1)

    if args.num_templates and not args.templates:
        logger.error("Number of templates provided without including templates")
        sys.exit(1)

    return args



import tempfile
from pathlib import Path

from abcfold.output.file_handlers import CifFile, NpyFile, NpzFile
from abcfold.output.utils import Af3Pae


def test_process_chai_output(test_data, output_objs):
    chai_output = output_objs.chai_output

    assert str(chai_output.output_dir) == str(test_data.test_chai1_6BJ9_)

    assert -1 in chai_output.output
    assert 0 in chai_output.output
    assert 1 in chai_output.output

    assert all(isinstance(pae_file, NpyFile) for pae_file in chai_output.pae_files)
    assert all(isinstance(cif_file, CifFile) for cif_file in chai_output.cif_files)
    assert all(
        isinstance(scores_file, NpzFile) for scores_file in chai_output.scores_files
    )


def test_chai_pae_to_af3_pae(output_objs):
    comparison_af3_output = output_objs.af3_output.af3_pae_files["seed-1"][0].data
    pae_file = output_objs.chai_output.pae_files[-1]
    for i, cif_file in enumerate(output_objs.chai_output.cif_files):
        assert cif_file.input_params
        pae = Af3Pae.from_chai1(
            pae_file.data[i],
            cif_file,
        )

        with tempfile.TemporaryDirectory() as temp_dir_str:
            temp_dir = Path(temp_dir_str)
            pae.to_file(temp_dir / "pae.json")
            assert (temp_dir / "pae.json").exists()

        # for some reason the lengths are different for atom - realted things
        # If it isn't breaking the output page generation, then it's fine
        assert len(pae.scores["pae"]) == len(comparison_af3_output["pae"])

        assert len(pae.scores["contact_probs"]) == len(
            comparison_af3_output["contact_probs"]
        )
        assert len(pae.scores["token_chain_ids"]) == len(
            comparison_af3_output["token_chain_ids"]
        )
        assert len(pae.scores["token_res_ids"]) == len(
            comparison_af3_output["token_res_ids"]
        )



import os
import tempfile

import pytest

try:
    import chai_lab  # noqa F401

    run_chai1 = True

except ImportError:
    run_chai1 = False


@pytest.mark.skipif(not run_chai1, reason="chai_lab not installed")
def test_generate_chai_command():
    from abcfold.chai1.run_chai1 import generate_chai_command

    input_fasta = "/road/to/nowhere.fasta"
    msa_dir = "/road/to/nowhere"
    with tempfile.NamedTemporaryFile(suffix=".csv", delete=False) as fp:
        constraints = fp.name
        output_dir = "/road/to/nowhere"

        cmd = generate_chai_command(
            input_fasta=input_fasta,
            msa_dir=msa_dir,
            input_constraints=constraints,
            output_dir=output_dir,
        )

    assert cmd[1].endswith("chai.py")
    assert "fold" in cmd
    assert input_fasta in cmd
    assert msa_dir in cmd
    assert constraints in cmd
    assert output_dir in cmd
    assert "--num-diffn-samples" in cmd
    assert "5" in cmd


@pytest.mark.skipif(
    os.getenv("CI") == "true" and not run_chai1,
    reason="Skipping test in CI environment",
)
def test_run_chai(test_data):
    pytest.importorskip("chai_lab")
    from abcfold.chai1.run_chai1 import run_chai

    with tempfile.TemporaryDirectory() as temp_dir:
        try:
            run_chai(
                test_data.test_inputA_json,
                temp_dir,
                save_input=True,
                test=True,
            )
        except Exception as e:
            print(e)
            assert False



import json
import logging
import random
import string
from pathlib import Path
from typing import Dict, List, Optional, Union

DELIM = "      "

logger = logging.getLogger("logger")


class BoltzYaml:
    """
    Object to convert an AlphaFold3 json file to a boltzmann yaml file.
    """

    def __init__(self, working_dir: Union[str, Path], create_files: bool = True):
        self.working_dir = working_dir
        self.yaml_string: str = ""
        self.msa_file: Optional[Union[str, Path]] = "null"
        self.__ids: List[Union[str, int]] = []
        self.__id_char: str = "A"
        self.__id_links: Dict[Union[str, int], list] = {}
        self.__create_files = create_files
        self.__non_ligands: List[str] = []
        self.__id_buffer: dict = {}

    @property
    def chain_ids(self) -> List[Union[str, int]]:
        return self.__ids

    @property
    def id_links(self) -> Dict[Union[str, int], list]:
        return self.__id_links

    def msa_to_file(self, msa: str, file_path: Union[str, Path]):
        """
        Takes a msa string and writes it to a file

        Args:
            msa (str): msa string
            file_path (Union[str, Path]): file path to write the msa to

        Returns:
            None
        """

        with open(file_path, "w") as f:
            f.write(msa)

    def json_to_yaml(
        self,
        json_file_or_dict: Union[dict, str, Path],
    ):
        """
        Main function to convert a json file or dict to a yaml string

        Args:
            json_file_or_dict (Union[dict, str, Path]): json file or dict

        Returns:
            str: a string representation of string
        """
        if isinstance(json_file_or_dict, str) or isinstance(json_file_or_dict, Path):
            with open(json_file_or_dict, "r") as f:
                json_dict = json.load(f)
        else:
            json_dict = json_file_or_dict

        self.get_ids(json_dict["sequences"])

        self.yaml_string = ""
        bonded_atom_string = ""

        self.yaml_string += self.add_version_number("1")
        for key, value in json_dict.items():
            if key == "sequences":
                if "sequences" not in self.yaml_string:
                    self.yaml_string += self.add_non_indented_string("sequences")
                for sequence_dict in value:
                    if any([key in sequence_dict for key in ["protein", "rna", "dna"]]):
                        self.yaml_string += self.sequence_to_yaml(sequence_dict)
                    if any([key in sequence_dict for key in ["ligand"]]):
                        self.yaml_string += self.add_ligand_information(
                            sequence_dict["ligand"]
                        )
            if key == "bondedAtomPairs" and isinstance(value, list):

                bonded_atom_string += self.bonded_atom_pairs_to_yaml(value)
                if "constraints" not in self.yaml_string and bonded_atom_string:
                    self.yaml_string += self.add_non_indented_string("constraints")
                self.yaml_string += bonded_atom_string

        return self.yaml_string

    def bonded_atom_pairs_to_yaml(self, bonded_atom_pairs: list):
        yaml_string = ""
        # counter = 0
        for pair in bonded_atom_pairs:

            if (pair[0][0] == pair[1][0]) and pair[0][1] not in self.__non_ligands:

                if pair[0][0] not in self.__id_links:
                    continue

                # I'm sorry
                if pair[0][0] not in self.__id_buffer:
                    self.__id_buffer[pair[0][0]] = 0
                else:
                    self.__id_buffer[pair[0][0]] += 1

                if self.__id_buffer[pair[0][0]] == 0:
                    first = pair[0][0]
                    second = self.__id_links[pair[0][0]][0]
                else:
                    first, second = (
                        self.__id_links[pair[0][0]][self.__id_buffer[pair[0][0]] - 1],
                        self.__id_links[pair[0][0]][self.__id_buffer[pair[0][0]]],
                    )
                if pair[0][1] < pair[1][1]:
                    pair[0] = [first, 1, pair[0][2]]
                    pair[1] = [second, 1, pair[1][2]]
                else:
                    pair[0] = [first, 1, pair[0][2]]
                    pair[1] = [second, 2, pair[1][2]]
            yaml_string += self.add_title("bond")
            yaml_string += self.add_key_and_value("atom1", pair[0])
            yaml_string += self.add_key_and_value("atom2", pair[1])

        return yaml_string

    def add_version_number(self, version: str):
        """
        Adds the version number to the yaml string

        Args:
            version (str): version number

        Returns:
            str: yaml string
        """
        return f"version: {version}\n"

    def add_non_indented_string(self, string: str):
        """
        Adds the sequence string to the yaml string

        Returns:
            str: yaml string
        """
        return f"{string}:\n"

    def add_id(self, id_: Union[str, list, int]):
        """
        Adds the id to the yaml string

        Args:
            id_ (Union[str, list, int]): id

        Returns:
            str: yaml string

        """

        if isinstance(id_, list):
            self.__ids.extend([id__ for id__ in id_ if id__ not in self.__ids])
            new_id = ", ".join([str(i).replace('"', "").replace("'", "") for i in id_])
        else:
            self.__ids.append(id_) if id_ not in self.__ids else None
            new_id = str(id_).replace('"', "").replace("'", "")

        return (
            f"{DELIM}{DELIM}id: {new_id}\n"
            if not isinstance(id_, list)
            else f"{DELIM}{DELIM}id: [{new_id}]\n"
        )

    def add_sequence(self, sequence: str):
        """
        Adds the sequence to the yaml string

        Args:
            sequence (str): sequence

        Returns:
            str: yaml string

        """
        return f"{DELIM}{DELIM}sequence: {sequence}\n"

    def add_msa(self, msa: Union[str, Path]):
        """
        Adds the msa file_path to the yaml string, double tabbed

        Args:
            msa (str): msa file_path

        Returns:
            str: yaml string
        """
        if not Path(msa).exists() and self.__create_files:
            msg = f"File {msa} does not exist"
            logger.critical(msg)
            raise FileNotFoundError()
        return f"{DELIM}{DELIM}msa: {msa}\n"

    def add_modifications(self, list_of_modifications: list):
        """
        Adds the modifications to the yaml string, double tabbed

        Args:
            list_of_modifications (list): list of modifications

        Returns:
            str: yaml string
        """
        yaml_string = ""
        yaml_string += f"{DELIM}{DELIM}modifications:\n"
        for modification in list_of_modifications:
            yaml_string += (
                f"{DELIM}{DELIM}{DELIM}- position: {modification['ptmPosition']}\n"
            )
            yaml_string += f"{DELIM}{DELIM}{DELIM}  ccd: {modification['ptmType']}\n"
        return yaml_string

    def add_key_and_value(self, key: str, value: str):
        """
        Adds the key and value to the yaml string, double tabbed

        Args:
            key (str): The key on the left of ':'
            value (str): The value on the right of ':'
        Returns:
            str: yaml string
        """
        return f"{DELIM}{DELIM}{key}: {value}\n"

    def add_ligand_information(self, ligand_dict: dict, linked_id=None):
        """
        Function to add ligand information to the yaml string

        Args:
            ligand_dict (dict): ligand dict

        Returns:
            str: yaml string
        """

        if "ccdCodes" in ligand_dict and len(ligand_dict["ccdCodes"]) == 0:
            return ""
        yaml_string = ""
        yaml_string += self.add_title("ligand")
        yaml_string += self.add_id(ligand_dict["id"])

        if "smiles" in ligand_dict:
            yaml_string += self.add_key_and_value("smiles", ligand_dict["smiles"])
        elif "ccdCodes" in ligand_dict:
            if isinstance(ligand_dict["ccdCodes"], str):
                yaml_string += self.add_key_and_value("ccd", ligand_dict["ccdCodes"])
            elif isinstance(ligand_dict["ccdCodes"], list):
                if linked_id is not None:

                    self.__add_linked_ids(linked_id, ligand_dict["id"])

                yaml_string += self.add_key_and_value("ccd", ligand_dict["ccdCodes"][0])

                yaml_string += self.add_ligand_information(
                    {
                        "id": self.find_next_id(),
                        "ccdCodes": ligand_dict["ccdCodes"][1:],
                    },
                    linked_id=ligand_dict["id"],
                )

        else:

            msg = "Ligand must have either a smiles or ccdCCodes"
            logger.critical(msg)
            raise ValueError()

        return yaml_string

    def add_sequence_information(self, sequence_dict: dict):
        """
        Adds the sequence information of protein, rna, dna to the yaml string

        Args:
            sequence_dict (dict): sequence dict
            msa_file (Union[str, Path]): msa file_path

        Returns:
            str: yaml string
        """
        yaml_string = ""
        yaml_string += self.add_id(sequence_dict["id"])
        yaml_string += (
            self.add_sequence(sequence_dict["sequence"])
            if "sequence" in sequence_dict
            else ""
        )
        if isinstance(sequence_dict["id"], str):
            id_ = [sequence_dict["id"]]
        else:
            id_ = sequence_dict["id"]

        self.__non_ligands.extend(id_)

        if self.msa_file is not None:
            (
                self.msa_to_file(sequence_dict["unpairedMsa"], self.msa_file)
                if self.__create_files
                else None
            )
            yaml_string += self.add_msa(self.msa_file)

        if "modifications" in sequence_dict and sequence_dict["modifications"]:
            yaml_string += self.add_modifications(sequence_dict["modifications"])

        return yaml_string

    def add_title(self, name: str):
        """
        Adds the title to the yaml string

        args:
            name (str): name of the title

        Returns:
            str: yaml string
        """
        return f"{DELIM}- {name}:\n"

    def sequence_to_yaml(self, sequence_dict: dict, yaml_string: str = ""):
        """
        Adds the sequence information to the yaml string

        Args:
            sequence_dict (dict): sequence dict
            yaml_string (str): yaml string

        Returns:
            str: yaml string
        """
        for sequence_type, sequence_info_dict in sequence_dict.items():
            yaml_string += self.add_title(sequence_type)
            self.msa_file = (
                (
                    Path(self.working_dir)
                    / f"{''.join(random.choices(string.ascii_letters, k=5))}.a3m"
                )
                if "unpairedMsa" in sequence_info_dict
                else None
            )

            yaml_string += self.add_sequence_information(sequence_info_dict)

        return yaml_string

    def write_yaml(self, file_path: Union[str, Path]):
        """
        Writes the yaml string to a file

        Args:
            file_path (Union[str, Path]): file path

        Returns:
            None
        """

        assert self.yaml_string, "No yaml string to write to file"
        assert Path(file_path).suffix == ".yaml", "File must have a .yaml extension"
        with open(file_path, "w") as f:
            f.write(self.yaml_string)

    def find_next_id(self):

        if self.__id_char not in self.__ids:
            return self.__id_char
        while self.__id_char in self.__ids:
            self.__id_char = chr(ord(self.__id_char) + 1)
        return self.__id_char

    def get_ids(self, sequences: list):
        for sequence in sequences:
            for key in sequence:
                for key2 in sequence[key]:
                    if key2 == "id":
                        if isinstance(sequence[key][key2], list):
                            self.__ids.extend(sequence[key][key2])
                            continue
                        self.__ids.append(sequence[key][key2])

    def __add_linked_ids(
        self, ligand_id: Union[str, int], linked_ligand_id: Union[str, int]
    ):
        if not self.__id_links:
            self.__id_links[ligand_id] = [linked_ligand_id]
            return
        for id_, value in self.__id_links.items():
            if ligand_id in value:
                self.__id_links[id_].append(linked_ligand_id)

                return



import logging
import subprocess
import sys

logger = logging.getLogger("logger")

BOLTZ_VERSION = "0.4.1"


def check_boltz1():
    try:
        import boltz as _  # noqa F401

        cmd = [sys.executable, "-m", "pip", "show", "boltz"]
        with subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as proc:
            stdout, stderr = proc.communicate()
            if proc.returncode != 0:
                raise subprocess.CalledProcessError(proc.returncode, cmd, stderr)

            version = None
            for line in stdout.decode().split("\n"):
                if line.startswith("Version:"):
                    version = line.split(":", 1)[1].strip()
                    break

            if version != BOLTZ_VERSION:
                raise ImportError(
                    f"Expected boltz version {BOLTZ_VERSION}, found {version}"
                )

    except (ImportError, ModuleNotFoundError):

        logger.info("Installing boltz package")
        cmd = [
            sys.executable,
            "-m",
            "pip",
            "install",
            f"boltz=={BOLTZ_VERSION}",
            "--no-cache-dir",
        ]

        logger.info("Running %s", " ".join(cmd))
        with subprocess.Popen(
            cmd,
            stdout=sys.stdout,
            stderr=subprocess.PIPE,
        ) as proc:
            proc.wait()
            if proc.returncode != 0:
                if proc.stderr:
                    logger.error(proc.stderr.read().decode())
                raise subprocess.CalledProcessError(proc.returncode, proc.args)



import os
from pathlib import Path

from jinja2 import Template

TEMPLATE = """
<!doctype html>
<html lang="en">
<head>
    <title>{{ title }}</title>
    <meta http-equiv="content-type" content="application/xhtml+xml; charset=utf-8"/>
    <meta name="viewport" content="width=device-width, minimum-scale=1.0, maximum-scale=1.0"/>
    <meta name="theme-color" content="#ffffff">

    <link rel="icon" type="image/svg+xml" href="{{ favicon_svg }}">
    <link rel="icon" sizes="any" href="{{ favicon_ico }}" >

    <link rel="stylesheet" type="text/css" href="{{ bootstrap_min_css }}"/>
    <link rel="stylesheet" type="text/css" href="{{ bootstrap_css }}"/>

    <link rel="stylesheet" type="text/css" href="{{ pae_viewer_standalone_layout_css }}"/>
    <link rel="stylesheet" type="text/css" href="{{ common_css }}"/>
    <link rel="stylesheet" type="text/css" href="{{ structure_carousel_css }}"/>
    <link rel="stylesheet" type="text/css" href="{{ structure_viewer_css }}"/>
    <link rel="stylesheet" type="text/css" href="{{ structure_panel_css }}"/>
    <link rel="stylesheet" type="text/css" href="{{ pae_viewer_css }}"/>
    <link rel="stylesheet" type="text/css" href="{{ complex_sequence_viewer_css }}"/>
    <link rel="stylesheet" type="text/css" href="{{ pae_viewer_standalone_overrides_css }}"/>

    <script type="text/javascript" src="{{ jquery_min_js }}"></script>
    <script type="text/javascript" src="{{ jquery_ui_min_js }}"></script>
    <script type="text/javascript" src="{{ bootstrap_bundle_min_js }}"></script>
    <script type="text/javascript" src="{{ ngl_js }}"></script>
    <script type="text/javascript" src="{{ chroma_min_js }}"></script>
    <script type="text/javascript" src="{{ file_saver_min_js }}"></script>

    <script defer src="{{ setup_js }}"></script>

    <script defer type="module" src="{{ structure_carousel_js }}"></script>
    <script defer type="module" src="{{ pae_viewer_js }}"></script>
    <script defer type="module" src="{{ complex_sequence_viewer_js }}"></script>
</head>
<body>

<input type="hidden" id="relativePath" value="{{ relative_path }}">

<noscript>
    <div class="alert alert-danger text-center">
        <strong>Error:</strong> JavaScript is disabled or not supported
        by your browser, but this page requires it to work.
    </div>
</noscript>
<div id="container">
    <div class="header" style="background-color: white">
        <img class="logo" src="../.feature_viewer/static/ABCfold-logo.svg" alt="ABCFold">
        <a href="https://github.com/rigdenlab/ABCFold.git" class="github-corner" aria-label="View source on Github"><svg width="80" height="80" viewBox="0 0 250 250" style="fill:#343635; color:#fff; float: right; border: 0;" aria-hidden="true"><path d="M0,0 L115,115 L130,115 L142,142 L250,250 L250,0 Z"></path><path d="M128.3,109.0 C113.8,99.7 119.0,89.6 119.0,89.6 C122.0,82.7 120.5,78.6 120.5,78.6 C119.2,72.0 123.4,76.3 123.4,76.3 C127.3,80.9 125.5,87.3 125.5,87.3 C122.9,97.6 130.6,101.9 134.4,103.2" fill="currentColor" style="transform-origin: 130px 106px;" class="octo-arm"></path><path d="M115.0,115.0 C114.9,115.1 118.7,116.5 119.8,115.4 L133.7,101.6 C136.9,99.2 139.9,98.4 142.2,98.6 C133.8,88.0 127.5,74.4 143.8,58.0 C148.5,53.4 154.0,51.2 159.7,51.0 C160.3,49.4 163.2,43.6 171.4,40.1 C171.4,40.1 176.1,42.5 178.8,56.2 C183.1,58.6 187.2,61.8 190.9,65.4 C194.5,69.0 197.7,73.2 200.1,77.6 C213.8,80.2 216.3,84.9 216.3,84.9 C212.7,93.1 206.9,96.0 205.4,96.6 C205.1,102.4 203.0,107.8 198.3,112.5 C181.9,128.9 168.3,122.5 157.7,114.1 C157.9,116.9 156.7,120.9 152.7,124.9 L141.0,136.5 C139.8,137.7 141.6,141.9 141.8,141.8 Z" fill="currentColor" class="octo-body"></path></svg></a><style>.github-corner:hover .octo-arm{animation:octocat-wave 560ms ease-in-out}@keyframes octocat-wave{0%,100%{transform:rotate(0)}20%,60%{transform:rotate(-25deg)}40%,80%{transform:rotate(10deg)}}@media (max-width:500px){.github-corner:hover .octo-arm{animation:none}.github-corner .octo-arm{animation:octocat-wave 560ms ease-in-out}}</style>
    </div>
    <main>
        {{ pae_viewer_demo_tpl}}
    </main>
    <footer class="footer">
    </footer>
</div>
</body>
</html>
"""  # noqa


def render_html(args, output_file):

    template = Template(TEMPLATE)
    # create a relative link between the current directory and the template

    relative_path = Path(os.path.relpath(args.src_path, Path(output_file).parent))

    rendered_html = template.render(
        title=args.title,
        favicon_svg=str(relative_path.joinpath("standalone", "favicon.svg")),
        favicon_ico=str(relative_path.joinpath("standalone", "favicon.ico")),
        bootstrap_min_css=str(
            relative_path.joinpath("standalone", "libs", "bootstrap.min.css")
        ),
        bootstrap_css=str(
            relative_path.joinpath("standalone", "libs", "bootstrap@5.1.3.css")
        ),
        pae_viewer_standalone_layout_css=str(
            relative_path.joinpath("standalone", "css", args.standalonecss)
        ),
        common_css=str(relative_path.joinpath("standalone", "css", "common.css")),
        structure_carousel_css=str(
            relative_path.joinpath("src", "css", "structure-carousel.css")
        ),
        structure_viewer_css=str(
            relative_path.joinpath("src", "css", "structure-viewer.css")
        ),
        structure_panel_css=str(
            relative_path.joinpath("src", "css", "structure-panel.css")
        ),
        pae_viewer_css=str(relative_path.joinpath("src", "css", "pae-viewer.css")),
        complex_sequence_viewer_css=str(
            relative_path.joinpath("src", "css", "complex-sequence-viewer.css")
        ),
        pae_viewer_standalone_overrides_css=str(
            relative_path.joinpath(
                "standalone", "css", "paeViewerStandaloneOverrides.css"
            )
        ),
        jquery_min_js=str(
            relative_path.joinpath("standalone", "libs", "jquery.min.js")
        ),
        jquery_ui_min_js=str(
            relative_path.joinpath("standalone", "libs", "jquery-ui.min.js")
        ),
        bootstrap_bundle_min_js=str(
            relative_path.joinpath("standalone", "libs", "bootstrap.bundle.min.js")
        ),
        ngl_js=str(relative_path.joinpath("standalone", "libs", "ngl.js")),
        chroma_min_js=str(
            relative_path.joinpath("standalone", "libs", "chroma.min.js")
        ),
        file_saver_min_js=str(
            relative_path.joinpath("standalone", "libs", "FileSaver.min.js")
        ),
        setup_js=str(relative_path.joinpath("standalone", "js", "setup.js")),
        structure_carousel_js=str(
            relative_path.joinpath("src", "modules", "structure-carousel.js")
        ),
        pae_viewer_js=str(relative_path.joinpath("src", "modules", "pae-viewer.js")),
        complex_sequence_viewer_js=str(
            relative_path.joinpath("src", "modules", "complex-sequence-viewer.js")
        ),
        header_title=args.title,
        header_subtitle="Interactive display of the Predicted Aligned Error",
        relative_path=relative_path,
        pae_viewer_demo_tpl="{{paeViewerDemo.tpl}}",
    )

    return rendered_html


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawTextHelpFormatter
    )

    parser.add_argument(
        "--output_file",
        help="Output file",
        type=str,
        default="pae-viewer.html",
    )

    parser.add_argument(
        "--title",
        help="Title of the HTML page",
        type=str,
        default="PAE Viewer",
    )
    parser.add_argument(
        "--standalonecss",
        type=str,
        default="paeViewerStandaloneLayout.css",
    )
    parser.add_argument(
        "--src_path",
        type=str,
        required=True,
    )

    args = parser.parse_args()
    rendered_html = render_html(args, output_file=args.output_file)

    # write html file
    with open(args.output_file, "w") as f:
        f.write(rendered_html)



import logging
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Union

from abcfold.boltz1.af3_to_boltz1 import BoltzYaml
from abcfold.boltz1.check_install import check_boltz1

logger = logging.getLogger("logger")


def run_boltz(
    input_json: Union[str, Path],
    output_dir: Union[str, Path],
    save_input: bool = False,
    test: bool = False,
    number_of_models: int = 5,
    num_recycles: int = 10,
):
    """
    Run Boltz1 using the input JSON file

    Args:
        input_json (Union[str, Path]): Path to the input JSON file
        output_dir (Union[str, Path]): Path to the output directory
        save_input (bool): If True, save the input yaml file and MSA to the output
        directory
        test (bool): If True, run the test command
        number_of_models (int): Number of models to generate

    Returns:
        None

    Raises:
        subprocess.CalledProcessError: If the Boltz1 command returns an error


    """
    input_json = Path(input_json)
    output_dir = Path(output_dir)

    logger.debug("Checking if boltz1 is installed")
    check_boltz1()

    with tempfile.TemporaryDirectory() as temp_dir:
        working_dir = Path(temp_dir)
        if save_input:
            logger.info("Saving input yaml file and msa to the output directory")
            working_dir = output_dir

        boltz_yaml = BoltzYaml(working_dir)
        boltz_yaml.json_to_yaml(input_json)
        out_file = working_dir.joinpath(f"{input_json.stem}.yaml")

        boltz_yaml.write_yaml(out_file)
        logger.info("Running Boltz1")
        cmd = (
            generate_boltz_command(out_file, output_dir, number_of_models, num_recycles)
            if not test
            else generate_boltz_test_command()
        )

        with subprocess.Popen(
            cmd,
            stdout=sys.stdout,
            stderr=subprocess.PIPE,
        ) as proc:
            _, stderr = proc.communicate()
            if proc.returncode != 0:
                if proc.stderr:
                    logger.error(stderr.decode())
                raise subprocess.CalledProcessError(proc.returncode, proc.args)

        logger.info("Boltz1 run complete")
        logger.info("Output files are in %s", output_dir)


def generate_boltz_command(
    input_yaml: Union[str, Path],
    output_dir: Union[str, Path],
    number_of_models: int = 5,
    num_recycles: int = 10,
) -> list:
    """
    Generate the Boltz1 command

    Args:
        input_yaml (Union[str, Path]): Path to the input YAML file
        output_dir (Union[str, Path]): Path to the output directory
        number_of_models (int): Number of models to generate

    Returns:
        list: The Boltz1 command
    """
    return [
        "boltz",
        "predict",
        str(input_yaml),
        "--out_dir",
        str(output_dir),
        "--override",
        "--write_full_pae",
        "--write_full_pde",
        "--diffusion_samples",
        str(number_of_models),
        "--recycling_steps",
        str(num_recycles),
    ]


def generate_boltz_test_command() -> list:
    """
    Generate the test command for Boltz1

    Args:
        None

    Returns:
        list: The Boltz1 test command
    """

    return [
        "boltz",
        "predict",
        "--help",
    ]



import os
import tempfile

import pytest

from abcfold.boltz1.run_boltz import generate_boltz_command, run_boltz


@pytest.mark.skipif(os.getenv("CI") == "true", reason="Skipping test in CI environment")
def test_run_boltz(test_data):

    with tempfile.TemporaryDirectory() as temp_dir:
        try:
            run_boltz(
                test_data.test_inputA_json,
                temp_dir,
                save_input=True,
                test=True,
            )
        except Exception as e:
            print(e)
            assert False


def test_generate_boltz_command():
    input_yaml = "/road/to/nowhere.yaml"
    output_dir = "/road/to/nowhere"

    cmd = generate_boltz_command(
        input_yaml=input_yaml,
        output_dir=output_dir,
    )

    assert "boltz" in cmd
    assert "predict" in cmd
    assert input_yaml in cmd
    assert output_dir in cmd
    assert "--override" in cmd



# Copyright (c) 2024 Chai Discovery, Inc.
# Licensed under the Apache License, Version 2.0.
# See the LICENSE file for details.

# Notice: This file has been modified to include a wrapper around the
# `run_inference` function to allow for the PAE (Predicted Aligned Error)
# to be output. The wrapper function captures the PAE output and
# integrates it into the command line interface.

"""Command line interface."""

import logging
from pathlib import Path

import numpy as np
import typer
from chai_lab.chai1 import run_inference

logging.basicConfig(level=logging.INFO)

CITATION = """
@article{Chai-1-Technical-Report,
    title        = {Chai-1: Decoding the molecular interactions of life},
    author       = {{Chai Discovery}},
    year         = 2024,
    journal      = {bioRxiv},
    publisher    = {Cold Spring Harbor Laboratory},
    doi          = {10.1101/2024.10.10.615955},
    url          = {https://www.biorxiv.org/content/early/2024/10/11/2024.10.10.615955},
    elocation-id = {2024.10.10.615955},
    eprint       = {https://www.biorxiv.org/content/early/2024/10/11/2024.10.1\
0.615955.full.pdf}
}
""".strip()


def citation():
    """Print citation information"""
    typer.echo(CITATION)


def run_inference_wrapper(
    fasta_file: Path,
    *,
    output_dir: Path,
    use_esm_embeddings: bool = True,
    use_msa_server: bool = False,
    msa_server_url: str = "https://api.colabfold.com",
    msa_directory: Path | None = None,
    constraint_path: Path | None = None,
    use_templates_server: bool = False,
    template_hits_path: Path | None = None,
    # Parameters controlling how we do inference
    recycle_msa_subsample: int = 0,
    num_trunk_recycles: int = 3,
    num_diffn_timesteps: int = 200,
    num_diffn_samples: int = 5,
    num_trunk_samples: int = 1,
    seed: int | None = None,
    device: str | None = None,
    low_memory: bool = True,
):

    result = run_inference(
        fasta_file=fasta_file,
        output_dir=output_dir,
        use_esm_embeddings=use_esm_embeddings,
        use_msa_server=use_msa_server,
        msa_server_url=msa_server_url,
        msa_directory=msa_directory,
        constraint_path=constraint_path,
        use_templates_server=use_templates_server,
        template_hits_path=template_hits_path,
        recycle_msa_subsample=recycle_msa_subsample,
        num_trunk_recycles=num_trunk_recycles,
        num_diffn_timesteps=num_diffn_timesteps,
        num_diffn_samples=num_diffn_samples,
        num_trunk_samples=num_trunk_samples,
        seed=seed,
        device=device,
        low_memory=low_memory,
    )

    np.save(f"{output_dir}/pae_scores.npy", result.pae)
    return result


def cli():
    app = typer.Typer()
    app.command("fold", help="Run Chai-1 to fold a complex.")(run_inference_wrapper)
    app.command("citation", help="Print citation information")(citation)
    app()


if __name__ == "__main__":
    cli()



import json
import shutil
import tempfile
from pathlib import Path

import pytest

from abcfold.scripts.abc_script_utils import get_custom_template
from abcfold.scripts.add_custom_template import main, run_custom_template


def test_get_custom_template(test_data):
    with open(test_data.test_inputA_json) as f:
        input_dict = json.load(f)
    get_custom_template(
        input_dict["sequences"][0],
        "A",
        test_data.test_6BJ9_cif,
        "B",
    )
    assert "templates" in input_dict["sequences"][0]["protein"]
    assert "queryIndices" in input_dict["sequences"][0]["protein"]["templates"][0]
    assert "templateIndices" in input_dict["sequences"][0]["protein"]["templates"][0]

    with open(test_data.test_inputA_json) as f:
        input_dict = json.load(f)

    get_custom_template(
        input_dict["sequences"][0],
        "B",
        test_data.test_6BJ9_cif,
        "A",
    )
    assert "templates" not in input_dict["sequences"][0]["protein"]

    with open(test_data.test_inputA_template_json) as f:
        input_dict = json.load(f)

    get_custom_template(
        input_dict["sequences"][0],
        "A",
        test_data.test_6BJ9_cif,
        "B",
    )
    assert "templates" in input_dict["sequences"][0]["protein"]

    get_custom_template(
        input_dict["sequences"][0],
        "A",
        test_data.test_6BJ9A_cif,
        None,
    )
    assert "templates" in input_dict["sequences"][0]["protein"]

    with open(test_data.test_inputAB_json) as f:
        input_dict = json.load(f)

    get_custom_template(
        input_dict["sequences"][0],
        "C",
        test_data.test_6BJ9_cif,
        "B",
    )

    assert "templates" not in input_dict["sequences"][0]["protein"]


def test_get_custom_template_errors(test_data):
    with open(test_data.test_inputA_json) as f:
        input_dict = json.load(f)

    with pytest.raises(FileNotFoundError):
        get_custom_template(
            input_dict["sequences"][0],
            "A",
            "road/to/nowhere",
            "C",
        )

    with pytest.raises(ValueError):
        get_custom_template(
            input_dict["sequences"][0],
            "A",
            test_data.test_6BJ9_cif,
            None,
        )

    with pytest.raises(ValueError):
        get_custom_template(
            input_dict["sequences"][0],
            "A",
            test_data.test_6BJ9_cif,
            "C",
        )


def test_run_custom_template(test_data):
    # run main function
    with pytest.raises(SystemExit):
        main()

    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        output_json = tmpdir / "output.json"
        run_custom_template(
            test_data.test_inputA_json,
            ["A"],
            [test_data.test_6BJ9_cif],
            ["B"],
            output_json=output_json,
            to_file=True,
        )
        shutil.copyfile(test_data.test_inputA_json, tmpdir / "input.json")

        run_custom_template(
            tmpdir / "input.json",
            ["A"],
            [test_data.test_6BJ9_cif],
            ["B"],
            output_json=None,
            to_file=True,
        )

        run_custom_template(
            test_data.test_inputRNA_json,
            ["A"],
            [test_data.test_6BJ9_cif],
            ["B"],
            output_json=output_json,
            to_file=True,
        )

    with pytest.raises(FileNotFoundError):
        run_custom_template(
            test_data.test_inputA_json,
            ["A"],
            ["road/to/nowhere"],
            ["C"],
            output_json=None,
            to_file=False,
        )
